# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass

from nemo.utils.megatron_utils import ApexGuardDefaults

try:
    from megatron.core.fusions.fused_bias_dropout import get_bias_dropout_add
    from megatron.core.fusions.fused_layer_norm import FusedLayerNorm
    from megatron.core.tensor_parallel.layers import ColumnParallelLinear, RowParallelLinear
    from megatron.core.transformer.attention import SelfAttention, SelfAttentionSubmodules
    from megatron.core.transformer.dot_product_attention import DotProductAttention
    from megatron.core.transformer.enums import AttnMaskType
    from megatron.core.transformer.identity_op import IdentityOp
    from megatron.core.transformer.mlp import MLP, MLPSubmodules
    from megatron.core.transformer.spec_utils import ModuleSpec, build_module
    from megatron.core.transformer.transformer_layer import TransformerLayer, TransformerLayerSubmodules
    from megatron.core.utils import make_viewless_tensor

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):
    TransformerConfig = ApexGuardDefaults
    HAVE_MEGATRON_CORE = False

try:
    from megatron.core.transformer.custom_layers.transformer_engine import (
        TEColumnParallelLinear,
        TEDotProductAttention,
        TENorm,
        TERowParallelLinear,
    )

    HAVE_TE = True
except (ImportError, ModuleNotFoundError):
    HAVE_TE = False


@dataclass
class TransformerLayerSubmodulesWithPostLNSupport(TransformerLayerSubmodules):
    """TransformerLayerSubmodules with post layer norm"""

    def __init__(self, post_att_layernorm, post_mlp_layernorm, **kwargs):
        super(TransformerLayerSubmodulesWithPostLNSupport, self).__init__(**kwargs)
        self.post_att_layernorm = post_att_layernorm
        self.post_mlp_layernorm = post_mlp_layernorm


class TransformerLayerWithPostLNSupport(TransformerLayer):
    """TransformerLayer with post layer norm."""

    def __init__(self, *args, **kwargs):
        super(TransformerLayerWithPostLNSupport, self).__init__(*args, **kwargs)
        # [Module add: Post attention LN]
        self.post_att_layernorm = build_module(
            self.submodules_config.post_att_layernorm,
            config=self.config,
            hidden_size=self.config.hidden_size,
            eps=self.config.layernorm_epsilon,
        )
        # [Module add: Post MLP LN]
        self.post_mlp_layernorm = build_module(
            self.submodules_config.post_mlp_layernorm,
            config=self.config,
            hidden_size=self.config.hidden_size,
            eps=self.config.layernorm_epsilon,
        )

    def forward(
        self,
        hidden_states,
        attention_mask,
        context=None,
        context_mask=None,
        rotary_pos_emb=None,
        rotary_pos_cos=None,
        rotary_pos_sin=None,
        attention_bias=None,
        inference_params=None,
        packed_seq_params=None,
        **kwargs,
    ):
        """Copy from megatron/core/transformer/transformer_layer.py with modification of applying
        extra post layer norm if needed."""
        # hidden_states: [s, b, h]

        # Residual connection.
        residual = hidden_states

        # Optional Input Layer norm
        input_layernorm_output = self.input_layernorm(hidden_states)

        # Self attention.
        attention_output_with_bias = self.self_attention(
            input_layernorm_output,
            attention_mask=attention_mask,
            inference_params=inference_params,
            rotary_pos_emb=rotary_pos_emb,
            packed_seq_params=packed_seq_params,
        )

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.self_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Post-LN after Self Attention
        hidden_states = self.post_att_layernorm(hidden_states)

        # Optional Layer norm after self-attention
        pre_cross_attn_layernorm_output = self.pre_cross_attn_layernorm(hidden_states)

        # Cross attention.
        attention_output_with_bias = self.cross_attention(
            pre_cross_attn_layernorm_output,
            attention_mask=context_mask,
            key_value_states=context,
            inference_params=inference_params,
        )

        if isinstance(attention_output_with_bias, dict) and "context" in attention_output_with_bias:
            context = attention_output_with_bias["context"]

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.cross_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Optional Layer norm post the cross-attention.
        pre_mlp_layernorm_output = self.pre_mlp_layernorm(hidden_states)

        # MLP.
        mlp_output_with_bias = self.mlp(pre_mlp_layernorm_output)

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.mlp_bda(self.training, self.config.bias_dropout_fusion)(
                mlp_output_with_bias, residual, self.hidden_dropout
            )

        # Post-LN after MLP
        hidden_states = self.post_mlp_layernorm(hidden_states)

        # Jit compiled function creates 'view' tensor. This tensor
        # potentially gets saved in the MPU checkpoint function context,
        # which rejects view tensors. While making a viewless tensor here
        # won't result in memory savings (like the data loader, or
        # p2p_communication), it serves to document the origin of this
        # 'view' tensor.
        output = make_viewless_tensor(inp=hidden_states, requires_grad=hidden_states.requires_grad, keep_graph=True)

        return output, context


# Use this spec to use lower level Transformer Engine modules (required for fp8 training)
def get_bert_layer_with_transformer_engine_spec_postln():
    """Retrieve the Layer Spec when using Transformer Engine"""
    return ModuleSpec(
        module=TransformerLayerWithPostLNSupport,
        submodules=TransformerLayerSubmodulesWithPostLNSupport(
            self_attention=ModuleSpec(
                module=SelfAttention,
                params={"attn_mask_type": AttnMaskType.padding},
                submodules=SelfAttentionSubmodules(
                    linear_qkv=TEColumnParallelLinear,
                    core_attention=TEDotProductAttention,
                    linear_proj=TERowParallelLinear,
                    q_layernorm=IdentityOp,
                    k_layernorm=IdentityOp,
                ),
            ),
            self_attn_bda=get_bias_dropout_add,
            post_att_layernorm=TENorm,
            mlp=ModuleSpec(
                module=MLP,
                submodules=MLPSubmodules(
                    linear_fc1=TEColumnParallelLinear,
                    linear_fc2=TERowParallelLinear,
                ),
            ),
            mlp_bda=get_bias_dropout_add,
            post_mlp_layernorm=TENorm,
        ),
    )


# Use this spec for an implementation using only modules in megatron core
def get_bert_layer_local_spec_postln():
    """Retrieve the Layer Spec when using MCore Engine"""
    return ModuleSpec(
        module=TransformerLayerWithPostLNSupport,
        submodules=TransformerLayerSubmodulesWithPostLNSupport(
            self_attention=ModuleSpec(
                module=SelfAttention,
                params={"attn_mask_type": AttnMaskType.padding},
                submodules=SelfAttentionSubmodules(
                    linear_qkv=ColumnParallelLinear,
                    core_attention=DotProductAttention,
                    linear_proj=RowParallelLinear,
                    q_layernorm=IdentityOp,
                    k_layernorm=IdentityOp,
                ),
            ),
            self_attn_bda=get_bias_dropout_add,
            post_att_layernorm=FusedLayerNorm,
            mlp=ModuleSpec(
                module=MLP,
                submodules=MLPSubmodules(
                    linear_fc1=ColumnParallelLinear,
                    linear_fc2=RowParallelLinear,
                ),
            ),
            mlp_bda=get_bias_dropout_add,
            post_mlp_layernorm=FusedLayerNorm,
        ),
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import logging
import os
import tarfile

from typing import IO, Union

LOGGER = logging.getLogger("NeMo")

try:
    from zarr.storage import BaseStore

    HAVE_ZARR = True
except Exception as e:
    LOGGER.warning(f"Cannot import zarr, support for zarr-based checkpoints is not available. {type(e).__name__}: {e}")
    BaseStore = object
    HAVE_ZARR = False


class TarPath:
    """
    A class that represents a path inside a TAR archive and behaves like pathlib.Path.

    Expected use is to create a TarPath for the root of the archive first, and then derive
    paths to other files or directories inside the archive like so:

    with TarPath('/path/to/archive.tar') as archive:
        myfile = archive / 'filename.txt'
        if myfile.exists():
            data = myfile.read()
            ...

    Only read and enumeration operations are supported.
    """

    def __init__(self, tar: Union[str, tarfile.TarFile, 'TarPath'], *parts):
        self._needs_to_close = False
        self._relpath = ''
        if isinstance(tar, TarPath):
            self._tar = tar._tar
            self._relpath = os.path.join(tar._relpath, *parts)
        elif isinstance(tar, tarfile.TarFile):
            self._tar = tar
            if parts:
                self._relpath = os.path.join(*parts)
        elif isinstance(tar, str):
            self._needs_to_close = True
            self._tar = tarfile.open(tar, 'r')
            if parts:
                self._relpath = os.path.join(*parts)
        else:
            raise ValueError(f"Unexpected argument type for TarPath: {type(tar).__name__}")

    def __del__(self):
        if self._needs_to_close:
            self._tar.close()

    def __truediv__(self, key) -> 'TarPath':
        return TarPath(self._tar, os.path.join(self._relpath, key))

    def __str__(self) -> str:
        return os.path.join(self._tar.name, self._relpath)

    @property
    def tarobject(self):
        """
        Returns the wrapped tar object.
        """
        return self._tar

    @property
    def relpath(self):
        """
        Returns the relative path of the path.
        """
        return self._relpath

    @property
    def name(self):
        """
        Returns the name of the path.
        """
        return os.path.split(self._relpath)[1]

    @property
    def suffix(self):
        """
        Returns the suffix of the path.
        """
        name = self.name
        i = name.rfind('.')
        if 0 < i < len(name) - 1:
            return name[i:]
        else:
            return ''

    def __enter__(self):
        self._tar.__enter__()
        return self

    def __exit__(self, *args):
        return self._tar.__exit__(*args)

    def exists(self):
        """
        Checks if the path exists.
        """
        try:
            self._tar.getmember(self._relpath)
            return True
        except KeyError:
            try:
                self._tar.getmember(os.path.join('.', self._relpath))
                return True
            except KeyError:
                return False

    def is_file(self):
        """
        Checks if the path is a file.
        """
        try:
            self._tar.getmember(self._relpath).isreg()
            return True
        except KeyError:
            try:
                self._tar.getmember(os.path.join('.', self._relpath)).isreg()
                return True
            except KeyError:
                return False

    def is_dir(self):
        """
        Checks if the path is a directory.
        """
        try:
            self._tar.getmember(self._relpath).isdir()
            return True
        except KeyError:
            try:
                self._tar.getmember(os.path.join('.', self._relpath)).isdir()
                return True
            except KeyError:
                return False

    def open(self, mode: str) -> IO[bytes]:
        """
        Opens a file in the archive.
        """
        if mode != 'r' and mode != 'rb':
            raise NotImplementedError()

        file = None
        try:
            # Try the relative path as-is first
            file = self._tar.extractfile(self._relpath)
        except KeyError:
            try:
                # Try the relative path with "./" prefix
                file = self._tar.extractfile(os.path.join('.', self._relpath))
            except KeyError:
                raise FileNotFoundError()

        if file is None:
            raise FileNotFoundError()

        return file

    def glob(self, pattern):
        """
        Returns an iterator over the files in the directory, matching the pattern.
        """
        for member in self._tar.getmembers():
            # Remove the "./" prefix, if any
            name = member.name[2:] if member.name.startswith('./') else member.name

            # If we're in a subdirectory, make sure the file is too, and remove that subdir component
            if self._relpath:
                if not name.startswith(self._relpath + '/'):
                    continue
                name = name[len(self._relpath) + 1 :]

            # See if the name matches the pattern
            if fnmatch.fnmatch(name, pattern):
                yield TarPath(self._tar, os.path.join(self._relpath, name))

    def rglob(self, pattern):
        """
        Returns an iterator over the files in the directory, including subdirectories.
        """
        for member in self._tar.getmembers():
            # Remove the "./" prefix, if any
            name = member.name[2:] if member.name.startswith('./') else member.name

            # If we're in a subdirectory, make sure the file is too, and remove that subdir component
            if self._relpath:
                if not name.startswith(self._relpath + '/'):
                    continue
                name = name[len(self._relpath) + 1 :]

            # See if any tail of the path matches the pattern, return full path if that's true
            parts = name.split('/')
            for i in range(len(parts)):
                subname = '/'.join(parts[i:])
                if fnmatch.fnmatch(subname, pattern):
                    yield TarPath(self._tar, os.path.join(self._relpath, name))
                    break

    def iterdir(self):
        """
        Returns an iterator over the files in the directory.
        """
        return self.glob('*')


class ZarrPathStore(BaseStore):
    """
    An implementation of read-only Store for zarr library
    that works with pathlib.Path or TarPath objects.
    """

    def __init__(self, tarpath: TarPath):
        assert HAVE_ZARR, "Package zarr>=2.18.2,<3.0.0 is required to use ZarrPathStore"
        self._path = tarpath
        self._writable = False
        self._erasable = False

    def __getitem__(self, key):
        with (self._path / key).open('rb') as file:
            return file.read()

    def __contains__(self, key):
        return (self._path / key).is_file()

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return sum(1 for _ in self.keys())

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def keys(self):
        """
        Returns an iterator over the keys in the store.
        """
        return self._path.iterdir()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

## NOTE: This script is present for github-actions testing only.
## There are no guarantees that this script is up-to-date with latest NeMo.

import argparse

import torch
from megatron.core.optimizer import OptimizerConfig
from pytorch_lightning.loggers import TensorBoardLogger
from transformers import AutoProcessor

from nemo import lightning as nl
from nemo.collections import llm, vlm
from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer
from nemo.collections.llm.api import train
from nemo.collections.multimodal.data.energon import ImageToken
from nemo.lightning import AutoResume, NeMoLogger
from nemo.lightning.pytorch.callbacks import ModelCheckpoint, ParameterDebugger
from nemo.lightning.pytorch.optim.megatron import MegatronOptimizerModule


def get_args():
    # pylint: disable=C0115,C0116
    parser = argparse.ArgumentParser(description='Train a small Llava Next model using NeMo 2.0')
    parser.add_argument('--devices', type=int, default=1, help="Number of devices to use for training")
    parser.add_argument('--max-steps', type=int, default=5, help="Number of steps to train for")
    parser.add_argument(
        '--experiment-dir', type=str, default=None, help="directory to write results and checkpoints to"
    )
    parser.add_argument(
        '--data-type',
        type=str,
        choices=['mock', 'energon'],
        default='mock',
        help="Type of data to use for training: mock or energon",
    )
    parser.add_argument(
        '--data-path',
        type=str,
        default=None,
        help="Path to the WebDataset for Energon data (only needed if data-type is energon)",
    )
    parser.add_argument(
        '--use-packed-sequence', action='store_true', help="Use packed sequence for more efficient training"
    )
    parser.add_argument('--gbs', type=int, default=2, help="Global batch size")
    parser.add_argument('--mbs', type=int, default=2, help="Micro batch size")
    parser.add_argument('--tensor-model-parallel-size', type=int, default=1, help="Tensor model parallel size")
    parser.add_argument('--pipeline-model-parallel-size', type=int, default=1, help="Pipeline model parallel size")
    parser.add_argument('--context-parallel-size', type=int, default=1, help="Context parallel size")

    return parser.parse_args()


if __name__ == '__main__':

    args = get_args()

    gbs = args.gbs
    mbs = args.mbs
    decoder_seq_length = 8192

    model_id = "llava-hf/llava-v1.6-vicuna-7b-hf"
    processor = AutoProcessor.from_pretrained(model_id)
    tokenizer = AutoTokenizer(model_id)

    # Setup data module based on type
    if args.data_type == 'mock':
        # Use mock data for simple testing
        data = vlm.LlavaNextMockDataModule(
            seq_length=decoder_seq_length,
            tokenizer=tokenizer,
            image_processor=processor.image_processor,
            global_batch_size=gbs,
            micro_batch_size=mbs,
            num_workers=0,
        )
    elif args.data_type == 'energon':
        # Validate args
        if not args.data_path:
            raise ValueError("For Energon data type, you must specify --data-path")

        from nemo.collections.multimodal.data.energon import EnergonMultiModalDataModule
        from nemo.collections.multimodal.data.energon.config import MultiModalSampleConfig
        from nemo.collections.vlm import LlavaNextTaskEncoder

        # Configure multimodal sample settings
        multimodal_sample_config = MultiModalSampleConfig(
            image_token=ImageToken(token_str="<image>", token_id=-200),
            ignore_place_holder=-100,
        )
        # Setting system prompt to empty string
        multimodal_sample_config.conversation_template_config.system = ''

        # Setup task encoder
        task_encoder = LlavaNextTaskEncoder(
            tokenizer=tokenizer.tokenizer,
            image_processor=processor.image_processor,
            multimodal_sample_config=multimodal_sample_config,
            packed_sequence=args.use_packed_sequence,
            packed_sequence_size=decoder_seq_length,
        )

        # Create data module with Energon
        data = EnergonMultiModalDataModule(
            path=args.data_path,
            tokenizer=tokenizer,
            image_processor=processor.image_processor,
            num_workers=2,
            micro_batch_size=mbs,
            global_batch_size=gbs,
            seq_length=decoder_seq_length,
            multimodal_sample_config=multimodal_sample_config,
            task_encoder=task_encoder,
            packing_buffer_size=200 if args.use_packed_sequence else None,
            virtual_epoch_length=10,  # Small value for testing
        )
    else:
        raise ValueError(f"Unknown data type: {args.data_type}")

    # Transformer configurations
    language_transformer_config = llm.Llama2Config7B(seq_length=decoder_seq_length, num_layers=2)

    vision_transformer_config = vlm.HFCLIPVisionConfig(
        pretrained_model_name_or_path="openai/clip-vit-large-patch14-336"
    )
    vision_projection_config = vlm.MultimodalProjectorConfig(
        projector_type="mlp2x_gelu",
        input_size=1024,
        hidden_size=4096,
        ffn_hidden_size=4096,
    )

    # Llava Next model configuration
    neva_config = vlm.LlavaNextConfig(
        language_transformer_config=language_transformer_config,
        vision_transformer_config=vision_transformer_config,
        vision_projection_config=vision_projection_config,
        freeze_language_model=True,
        freeze_vision_model=True,
    )

    model = vlm.LlavaNextModel(neva_config, tokenizer=data.tokenizer)

    strategy = nl.MegatronStrategy(
        tensor_model_parallel_size=args.tensor_model_parallel_size,
        pipeline_model_parallel_size=args.pipeline_model_parallel_size,
        context_parallel_size=args.context_parallel_size,
        encoder_pipeline_model_parallel_size=0,
        pipeline_dtype=torch.bfloat16,
    )
    checkpoint_callback = ModelCheckpoint(
        every_n_train_steps=5000,
        save_optim_on_train_end=True,
    )

    def create_verify_precision(precision: torch.dtype):
        def verify_precision(tensor: torch.Tensor) -> None:
            assert tensor.dtype == precision

        return verify_precision

    debugger = ParameterDebugger(
        param_fn=create_verify_precision(torch.bfloat16),
        grad_fn=create_verify_precision(torch.float32),
        log_on_hooks=["on_train_start", "on_train_end"],
    )
    callbacks = [checkpoint_callback, debugger]

    loggers = []
    tensorboard_logger = TensorBoardLogger(
        save_dir='dummy',  ## NOTE: this gets overwritten by default
    )
    loggers.append(tensorboard_logger)

    opt_config = OptimizerConfig(
        optimizer='adam',
        lr=6e-4,
        min_lr=6e-5,
        use_distributed_optimizer=False,
        bf16=True,
    )
    opt = MegatronOptimizerModule(config=opt_config)

    trainer = nl.Trainer(
        devices=args.devices,
        max_steps=args.max_steps,
        accelerator="gpu",
        strategy=strategy,
        logger=loggers,
        callbacks=callbacks,
        log_every_n_steps=1,
        limit_val_batches=2,
        plugins=nl.MegatronMixedPrecision(precision="bf16-mixed"),
    )

    nemo_logger = NeMoLogger(
        log_dir=args.experiment_dir,
    )

    resume = AutoResume(
        resume_if_exists=True,
        resume_ignore_no_checkpoint=True,
    )

    train(
        model=model,
        data=data,
        trainer=trainer,
        log=nemo_logger,
        resume=resume,
        tokenizer='data',
        optim=opt,
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable

import torch

from nemo.collections.asr.inference.streaming.decoders.greedy.greedy_decoder import GreedyDecoder


class RNNTGreedyDecoder(GreedyDecoder):
    """RNNT Greedy decoder class"""

    def __init__(self, vocabulary: list[str], conf_func: Callable = None):
        """
        Initialize the RNNTGreedyDecoder
        Args:
            vocabulary (list[str]): list of vocabulary tokens
            conf_func (Callable): function to compute confidence
        """
        super().__init__(vocabulary, conf_func)

    def __call__(
        self,
        global_timestamps: torch.Tensor | list[int],
        tokens: torch.Tensor | list[int],
        length: int,
        offset: int = 0,
    ) -> tuple[dict, list[int], int]:
        """
        Decode the RNNT hypothesis using timestamps
        Args:
            global_timestamps (torch.Tensor | list[int]): global timestamps since the start of the stream
            tokens (torch.Tensor | list[int]): tokens since the start of the stream
            length (int): length of the alignment
            offset (int): offset to apply to the timestamps to make them local
        Returns:
            tuple[dict, list[int], int]:
                output: dictionary containing the decoded tokens, timestamps, and confidences
                current labels: list of current labels including the blank token
                new offset: new offset value for the next decoding step
        """
        if isinstance(global_timestamps, list):
            global_timestamps = torch.tensor(global_timestamps)
        if isinstance(tokens, list):
            tokens = torch.tensor(tokens)

        output = {"tokens": [], "timesteps": [], "confidences": [], "last_token": None, "last_token_idx": None}
        cur_labels = [self.blank_id] * length
        new_offset = len(tokens)
        if offset > 0:
            trimmed_tokens = tokens[offset:].tolist()
            trimmed_timestamps = global_timestamps[offset:].tolist()
        else:
            trimmed_tokens = tokens.tolist()
            trimmed_timestamps = global_timestamps.tolist()

        if len(trimmed_tokens) == 0:
            return output, cur_labels, new_offset

        output["tokens"].extend(trimmed_tokens)
        output["timesteps"].extend(trimmed_timestamps)
        output["confidences"].extend([0.0] * len(trimmed_tokens))
        output["last_token"] = trimmed_tokens[-1]
        output["last_token_idx"] = trimmed_timestamps[-1]

        for t, token in zip(trimmed_timestamps, trimmed_tokens):
            cur_labels[t % length] = token
        return output, cur_labels, new_offset


class ClippedRNNTGreedyDecoder:
    """
    Clipped RNNT Greedy decoder class
    Decodes the tokens within a given clip range and returns the clipped tokens and timestamps.
    """

    def __init__(self, vocabulary: list[str], tokens_per_frame: int, conf_func: Callable = None, endpointer=None):
        """
        Initialize the ClippedRNNTGreedyDecoder
        Args:
            vocabulary (list[str]): list of vocabulary tokens
            tokens_per_frame (int): number of tokens per frame
            conf_func (Callable): function to compute confidence
            endpointer (Any): endpointer to detect EOU
        """
        self.greedy_decoder = RNNTGreedyDecoder(vocabulary, conf_func)
        self.endpointer = endpointer
        self.tokens_per_frame = tokens_per_frame

    @staticmethod
    def extract_clipped_and_tail_single_pass(
        timesteps: torch.Tensor, tokens: torch.Tensor, start_idx: int, end_idx: int, return_tail_result: bool
    ) -> tuple[list[int], list[int], list[int]]:
        """
        Extract clipped and tail data using tensor operations - no conversion overhead
        """
        if len(timesteps) == 0:
            return [], [], []
        clipped_mask = (timesteps >= start_idx) & (timesteps < end_idx)
        clipped_timesteps = timesteps[clipped_mask].tolist()
        clipped_tokens = tokens[clipped_mask].tolist()
        tail_tokens = []
        if return_tail_result:
            tail_mask = timesteps >= end_idx
            if tail_mask.any():
                tail_tokens = tokens[tail_mask].tolist()

        return clipped_timesteps, clipped_tokens, tail_tokens

    def __call__(
        self,
        global_timesteps: torch.Tensor,
        tokens: torch.Tensor,
        clip_start: int,
        clip_end: int,
        alignment_length: int,
        is_last: bool = False,
        is_start: bool = True,
        return_tail_result: bool = False,
        state_start_idx: int = 0,
        state_end_idx: int = 0,
        timestamp_offset: int = 0,
        vad_segments: torch.Tensor = None,
        stop_history_eou: int = None,
    ) -> tuple[dict, dict, bool, int, int]:
        """
        Decode using timestamps instead of dense alignment
        Optimized version with vectorized operations and single-pass processing
        Args:
            global_timesteps (torch.Tensor): global timestamps since the start of the stream
            tokens (torch.Tensor): tokens
            clip_start (int): start index of the clip
            clip_end (int): end index of the clip
            alignment_length (int): length of the alignment
            is_last (bool): is the last frame or not.
            is_start (bool): is the first frame for this stream or not.
            return_tail_result (bool): return tail result left after clip_end in the buffer
            state_start_idx (int): start index from stream state
            state_end_idx (int): end index from stream state
            timestamp_offset (int): offset to apply to the timestamps to make them local
            vad_segments (torch.Tensor): Optional VAD segments to use for end-of-utterance detection
            stop_history_eou (int): stop history of EOU, if None then use the default stop history
        Returns:
            tuple[dict, dict, bool, int, int]:
                clipped output, tail output, is_eou, updated start_idx, updated end_idx
        """
        # Initialize end-of-utterance state based on input parameters
        if timestamp_offset:
            timesteps = global_timesteps - timestamp_offset
        else:
            timesteps = global_timesteps
        is_eou = is_last
        eou_detected_at = alignment_length
        start_idx, end_idx = state_start_idx, state_end_idx
        if end_idx > clip_start:
            end_idx -= self.tokens_per_frame
            start_idx = end_idx
        if is_start:
            start_idx, end_idx = clip_start, clip_start
        elif end_idx <= clip_start:
            start_idx, end_idx = clip_start, clip_end

        if len(timesteps) == 0 or len(tokens) == 0:
            return (
                {"tokens": [], "timesteps": [], "confidences": [], "last_token": None, "last_token_idx": None},
                {"tokens": []},
                True,
                start_idx,
                end_idx,
            )

        mask = timesteps >= start_idx
        timesteps_trimmed = timesteps[mask]
        tokens_trimmed = tokens[mask]
        # If not already at end of utterance and endpointer exists, try to detect end of utterance
        if not is_eou and self.endpointer is not None:
            if vad_segments is not None and len(vad_segments) > 0:
                if vad_segments[-1][1] != 0.0:
                    is_eou, eou_detected_at = self.endpointer.detect_eou_vad(
                        vad_segments=vad_segments, search_start_point=start_idx, stop_history_eou=stop_history_eou
                    )
                else:
                    is_eou = True
                    eou_detected_at = -1
            else:
                is_eou, eou_detected_at = self.endpointer.detect_eou_given_timestamps(
                    timesteps=timesteps_trimmed,
                    tokens=tokens_trimmed,
                    alignment_length=alignment_length,
                    stop_history_eou=stop_history_eou,
                )
        # If EOU is detected beyond current end frame, extend end frame to include it
        if is_eou and eou_detected_at > end_idx:
            end_idx = min(eou_detected_at, alignment_length)

        # If the end frame is within the clip range, set the end frame to the clip end
        if clip_start <= end_idx < clip_end:
            end_idx = clip_end
            is_eou = False
        clipped_timesteps, clipped_tokens, tail_tokens = self.extract_clipped_and_tail_single_pass(
            timesteps, tokens, start_idx, end_idx, return_tail_result
        )
        # Make timestamps global again
        if timestamp_offset:
            clipped_timesteps = [t + timestamp_offset for t in clipped_timesteps]
        # Initialize output with last_token tracking like in __call__ method
        clipped_output = {
            "tokens": clipped_tokens,
            "timesteps": clipped_timesteps,
            "confidences": [0.0] * len(clipped_tokens) if len(clipped_tokens) > 0 else [],
            "last_token": None,
            "last_token_idx": None,
        }

        # Set last_token and last_token_idx if there are tokens
        if len(clipped_tokens) > 0:
            clipped_output["last_token"] = clipped_tokens[-1]
            clipped_output["last_token_idx"] = clipped_timesteps[-1] if len(clipped_timesteps) > 0 else None

        # Create tail output
        tail_output = {"tokens": tail_tokens}
        return clipped_output, tail_output, is_eou, start_idx, end_idx



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from typing import Callable

import nemo_run as run
import numpy as np
import pandas as pd
import torch
import torch.distributed as dist
import webdataset as wds
from einops import rearrange
from transformers import T5EncoderModel, T5TokenizerFast

from nemo.collections.common.video_tokenizers.cosmos_tokenizer import CausalVideoTokenizer
from nemo.collections.common.video_tokenizers.utils import read_image, resize_video


def initialize_text_encoder(t5_cache_dir):
    """
    Initializes the T5 tokenizer and encoder model, loading them from a specified cache directory.

    Args:
        t5_cache_dir (str): Path to the cache directory for storing the pretrained model files.

    Returns:
        tuple: A tuple containing the tokenizer and encoder model instances.
    """

    # Load tokenizer and text encoder, save in cache directory
    tokenizer = T5TokenizerFast.from_pretrained("google-t5/t5-11b", cache_dir=t5_cache_dir)
    text_encoder = T5EncoderModel.from_pretrained("google-t5/t5-11b", cache_dir=t5_cache_dir)
    text_encoder.to("cuda")
    text_encoder.eval()

    return tokenizer, text_encoder


# Load dataset from HuggingFace
df = pd.read_parquet("hf://datasets/huggan/smithsonian_butterflies_subset/data/train-00000-of-00001.parquet")
# Load Cosmos tokenizer from HuggingFace
autoencoder = CausalVideoTokenizer.from_pretrained("CosmosCausalCV_f4x8x8")
# Load T5-XXL text encoder
t5_cache_dir = ''  # Use your own custom cache path
tokenizer, text_encoder = initialize_text_encoder(t5_cache_dir)


class EncodedSample:
    """
    A class representing an encoded sample, containing the text encoding, length,
    attention mask, and offset mappings.

    Attributes:
        encoded_text (np.ndarray): Encoded text array.
        length (int): Length of the encoding.
        attn_mask (np.ndarray): Attention mask for the encoding.
        offset_mappings (np.ndarray): Mappings for offset positions.
    """

    def __init__(self, encoded_text: np.ndarray, length: int, attn_mask: np.ndarray, offset_mappings: np.ndarray):
        self.encoded_text = encoded_text
        self.length = length
        self.attn_mask = attn_mask
        self.offset_mappings = offset_mappings

    def truncate(self) -> None:
        """
        Truncates the encoded text, attention mask, and offset mappings to the specified length.
        """
        self.encoded_text = self.encoded_text[0 : self.length].astype(np.float16)
        self.attn_mask = self.attn_mask[0 : self.length].astype(np.int32)
        if self.offset_mappings is not None:
            self.offset_mappings = self.offset_mappings[0 : self.length].astype(np.int32)


@torch.no_grad()
def encode_for_batch(
    tokenizer, encoder, prompts: list[str], truncate: bool = True, max_length=512, output_mapping=True
):
    """
    Encodes a batch of text prompts into T5 embeddings.

    Args:
        tokenizer: Tokenizer instance for encoding.
        encoder: T5 encoder model instance.
        prompts (list[str]): List of text prompts to encode.
        truncate (bool): If True, truncates the output embeddings.
        max_length (int): Maximum length for each encoded prompt.
        output_mapping (bool): If True, returns offset mappings for each prompt.

    Returns:
        list[EncodedSample]: A list of encoded samples containing text encodings and masks.
    """
    batch_encoding = tokenizer.batch_encode_plus(
        prompts,
        return_tensors="pt",
        truncation=True,
        padding="max_length",
        max_length=max_length,
        return_length=True,
        return_offsets_mapping=output_mapping,
    )

    # We expect all the processing is done in GPU.
    input_ids = batch_encoding.input_ids.cuda()
    attn_mask = batch_encoding.attention_mask.cuda()
    if output_mapping:
        offsets_mapping = batch_encoding["offset_mapping"]
        offsets_mapping = offsets_mapping.cpu().numpy()
    else:
        offsets_mapping = None

    outputs = encoder(input_ids=input_ids, attention_mask=attn_mask)  # type: ignore
    encoded_text = outputs.last_hidden_state

    lengths = attn_mask.sum(dim=1).cpu()
    for batch_id in range(encoded_text.shape[0]):
        encoded_text[batch_id][lengths[batch_id] :] = 0

    encoded_text = encoded_text.cpu().numpy()
    attn_mask = attn_mask.cpu().numpy()

    encoded_text = encoded_text[:, :max_length]
    attn_mask = attn_mask[:, :max_length]

    out = []
    for idx in range(encoded_text.shape[0]):
        if output_mapping:
            offsets = offsets_mapping[idx]
        else:
            offsets = None

        out.append(EncodedSample(encoded_text[idx].astype(np.float16), lengths[idx], attn_mask[idx], offsets))
    if truncate:
        for x in out:
            x.truncate()
    return out


def generate_t5_embed(tokenizer, text_encoder, prompt, t5_embeding_max_length=512):
    """
    Generates a T5 embedding for a single text prompt.

    Args:
        tokenizer: T5 tokenizer instance.
        text_encoder: T5 encoder model instance.
        prompt (str): The text prompt to encode.
        t5_embeding_max_length (int): Maximum length for the embedding.

    Returns:
        torch.Tensor: Padded T5 embedding tensor.
    """
    # encode text to t5 embedding
    out = encode_for_batch(tokenizer, text_encoder, [prompt])[0]
    encoded_text = torch.tensor(out.encoded_text, dtype=torch.bfloat16)

    # padding t5 embedding to t5_embeding_max_length
    L, C = encoded_text.shape
    t5_embed = torch.zeros(1, t5_embeding_max_length, C, dtype=torch.bfloat16)
    t5_embed[0, :L] = encoded_text

    return t5_embed


def get_start_end_idx_for_this_rank(dataset_size, rank, world_size):
    """
    Calculates the start and end indices for distributed processing based on rank.

    Args:
        dataset_size (int): Total dataset size.
        rank (int): Current process rank.
        world_size (int): Total number of processes.

    Returns:
        tuple: (start index, end index) for the rank.
    """
    split_size = dataset_size // world_size
    start_idx = rank * split_size
    # The last rank takes the remainder
    end_idx = start_idx + split_size if rank != world_size - 1 else dataset_size
    return start_idx, end_idx


def butterfly_process_func(index):
    """
    Generates a sample dictionary with image latent tensor, caption, and metadata.

    Args:
        index (int): Index of the dataset row.

    Returns:
        dict: Dictionary containing processed image latents, embeddings, and metadata.
    """
    # Access the data from the dataframe
    row = df.iloc[index]
    image_url = row["image_url"]
    image_caption = row["name"]

    # Process image
    video = read_image(image_url)
    video = rearrange(video, 'h w (t c) -> t h w c', t=1)
    video = resize_video(video, short_size=512)
    batch_video = video[np.newaxis, ...]

    # Run autoencoder to get latents
    _, image_latent = autoencoder(batch_video, temporal_window=1)

    text_embedding = generate_t5_embed(tokenizer, text_encoder, image_caption)

    # Construct sample dictionary
    sample = {
        "__key__": f"{index:06}",
        ".pth": image_latent.to(dtype=torch.bfloat16),
        ".pickle": pickle.dumps(text_embedding),
        ".json": {
            "image_height": batch_video.shape[2],
            "image_width": batch_video.shape[3],
            # Add additional score as metadata
        },
    }
    return sample


@torch.no_grad()
@run.cli.entrypoint
def prepare(process_func: Callable, output_dir: str = 'output'):
    """
    Prepares a WebDataset using the specified processing function, for distributed settings.

    Args:
        process_func (Callable): Function to process each dataset entry.
        output_dir (str): Output directory to save processed dataset.

    """
    rank = dist.get_rank()
    world_size = torch.distributed.get_world_size()

    start_idx, end_idx = get_start_end_idx_for_this_rank(len(df), rank, world_size)
    os.makedirs(output_dir, exist_ok=True)
    output_tar = os.path.join(output_dir, f"rank{rank}-%06d.tar")

    with wds.ShardWriter(output_tar, maxcount=10000) as sink:
        for i in range(start_idx, end_idx):
            sample = process_func(i)
            # Write sample to tar file
            sink.write(sample)


@run.cli.factory(target=prepare)
def prepare_butterfly_dataset() -> run.Partial:
    """
    Prepares the butterfly dataset for distributed training.

    Returns:
        run.Partial: Partially configured run for WebDataset preparation.
    """
    recipe = run.Partial(prepare, process_func=butterfly_process_func, output_dir='butterfly_webdataset')
    return recipe


if __name__ == '__main__':
    dist.init_process_group("nccl")
    local_rank = int(os.environ['LOCAL_RANK'])
    torch.cuda.set_device(local_rank)
    run.cli.main(prepare, default_factory=prepare_butterfly_dataset)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

from nemo.export.tarutils import TarPath

CONFIG_NAME = "config.json"
WEIGHTS_NAME = "rank{}.safetensors"


def is_qnemo_checkpoint(path: str) -> bool:
    """Detect if a given path is a TensorRT-LLM a.k.a. "qnemo" checkpoint based on config & tensor data presence."""
    if os.path.isdir(path):
        path = Path(path)
    else:
        path = TarPath(path)
    config_path = path / CONFIG_NAME
    tensor_path = path / WEIGHTS_NAME.format(0)
    return config_path.exists() and tensor_path.exists()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
# pylint: skip-file

import copy
import io
import math
import os
from typing import Dict, List, Optional, Union

import numpy as np
import torch
from megatron.core import parallel_state
from omegaconf import DictConfig, ListConfig, open_dict

from nemo.collections.asr.data.audio_to_text import (
    VALID_FILE_FORMATS,
    cache_datastore_manifests,
    expand_sharded_filepaths,
    shard_manifests_if_needed,
)
from nemo.collections.asr.data.audio_to_text_dataset import convert_to_config_list, get_chain_dataset
from nemo.collections.asr.parts.preprocessing.features import WaveformFeaturizer
from nemo.collections.asr.parts.preprocessing.segment import ChannelSelectorType
from nemo.collections.common.data.blendable_dataset import BlendableDataset
from nemo.collections.common.data.dataset import ConcatDataset
from nemo.collections.common.parts.preprocessing import collections
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import (
    TextProcessing,
    build_loss_mask,
    ceil_to_nearest,
    get_num_samples_from_files,
    maybe_cast_to_list,
)
from nemo.core.classes import Dataset, IterableDataset
from nemo.utils import logging, logging_mode
from nemo.utils import webdataset as wds
from nemo.utils.distributed import webdataset_split_by_workers

__all__ = [
    'AudioTextDataset',
    'TarredAudioTextDataset',
    'get_tarred_audio_text_dataset_from_config',
    'get_audio_text_dataset_from_config',
]


def _audio_collate_fn(audio_signals, audio_lengths):
    """collate batch of audio sig, audio len, tokens, tokens len
    Args:
        audio_signals: List[Tensor]
        audio_lengths: List[Tensor]
    """

    max_audio_len = 0
    has_audio = audio_lengths[0] is not None
    if has_audio:
        max_audio_len = max(audio_lengths).item()

    audio_signals_padded = []
    for sig, sig_len in zip(audio_signals, audio_lengths):
        if has_audio:
            sig_len = sig_len.item()
            if sig_len < max_audio_len:
                pad = (0, max_audio_len - sig_len)
                sig = torch.nn.functional.pad(sig, pad)
            audio_signals_padded.append(sig)

    if has_audio:
        audio_signals_padded = torch.stack(audio_signals_padded)
        audio_lengths = torch.stack(audio_lengths)
    else:
        audio_signals_padded, audio_lengths = None, None

    return audio_signals_padded, audio_lengths


def _collate_item(item: Union[torch.Tensor, np.ndarray, List], max_length: int, pad_id: int = 0):
    # function copied from nemo/collections/nlp/data/language_modelling/megatron/gpt_sft_dataset.py
    item = maybe_cast_to_list(item)
    # max_length = max([len(x) for x in item]) if item else 0
    # here [0] should be tokenizer.pad_id
    item = [x + [pad_id] * (max_length - len(x)) for x in item]
    return item


def _speechllm_audio_text_collate_fn(
    batch: Dict,
    tokens_to_generate: int,
    pad_to_max_length: bool,
    max_seq_length: int,
    text_pad_id: int,
):
    sample_ids = [x["idx"] for x in batch]
    sample_ids = torch.tensor(sample_ids, dtype=torch.int32)

    audio_signal = [x["audio_signal"] for x in batch]
    audio_lengths = [x["audio_length"] for x in batch]
    audio_signal, audio_lengths = _audio_collate_fn(audio_signal, audio_lengths)

    input_ids = [item['input_ids'][:-1] for item in batch]
    labels = [item['input_ids'][1:] for item in batch]
    contexts = [item['context_ids'] for item in batch]
    context_lengths = torch.LongTensor([item['context_length'] for item in batch])
    answers = [item['answer_ids'] for item in batch]

    loss_mask = [build_loss_mask(item)[1:] for item in batch]

    max_length = max([len(x) for x in input_ids]) + tokens_to_generate
    # increase max length to nearest multiple of 4 or 8
    if pad_to_max_length:
        max_length = max_seq_length
    else:
        max_length = min(max_seq_length, ceil_to_nearest(max_length, 8))
    assert max_length <= max_seq_length

    position_ids = [list(range(max_length)) for _ in batch]
    position_ids = torch.LongTensor(position_ids)
    input_ids = torch.LongTensor(_collate_item(input_ids, max_length=max_length, pad_id=text_pad_id))
    input_length = torch.LongTensor([len(x) for x in input_ids])
    labels = torch.LongTensor(_collate_item(labels, max_length=max_length, pad_id=text_pad_id))
    loss_mask = torch.LongTensor(_collate_item(loss_mask, max_length=max_length, pad_id=0))
    contexts = torch.LongTensor(_collate_item(contexts, max_length=max_length, pad_id=text_pad_id))
    answers = torch.LongTensor(_collate_item(answers, max_length=max_length, pad_id=text_pad_id))

    batch = {
        'sample_ids': sample_ids,
        'audio_signal': audio_signal,
        'audio_signal_length': audio_lengths,
        'tokens': input_ids,
        'tokens_length': input_length,
        'labels': labels,
        'loss_mask': loss_mask,
        'position_ids': position_ids,
        'contexts': contexts,
        'context_lengths': context_lengths,
        'answers': answers,
        'max_length': torch.LongTensor(max_length),
        'metadata': [x['metadata'] for x in batch],
    }

    return batch


def _speechllm_multi_audio_text_collate_fn(
    batch: Dict,
    tokens_to_generate: int,
    pad_to_max_length: bool,
    max_seq_length: int,
    text_pad_id: int,
):
    """Collate function for multi audio case."""
    context_start_idx = [item['context_start_idx'] for item in batch]

    audio_signals = [x["audio_signal"] for x in batch]
    audio_lengths = [x["audio_length"] for x in batch]
    num_audios = [len(x) for x in audio_signals]

    # put all audios from all samples in one batch
    audio_signals_merged = [item for audio_list in audio_signals for item in audio_list]
    audio_lengths_merged = [item for length_list in audio_lengths for item in length_list]
    audio_signals_merged, audio_lengths_merged = _audio_collate_fn(audio_signals_merged, audio_lengths_merged)

    for i in range(len(batch)):
        # create dummy audio_signal and audio_length for _speechllm_audio_text_collate_fn()
        batch[i]["audio_signal"] = audio_signals[i][0]
        batch[i]["audio_length"] = audio_lengths[i][0]

    batch = _speechllm_audio_text_collate_fn(batch, tokens_to_generate, pad_to_max_length, max_seq_length, text_pad_id)

    # add multi audio specific fields
    batch['context_start_idx'] = list(context_start_idx)
    batch['num_audios'] = torch.LongTensor(num_audios)
    batch['audio_signal'] = audio_signals_merged
    batch['audio_signal_length'] = audio_lengths_merged

    return batch


class AudioTextDataset(Dataset):
    """
    Dataset that loads tensors via a json file containing paths to audio files, transcripts,
    and durations (in seconds).
    Each new line is a different sample. Example below:

    .. code-block:: json

        {"audio_filepath": "1.wav", "duration": 1.12, "question": "what is the capital of France?", "answer": "Paris"}
        {"audio_filepath": "2.wav", "duration": 2.15, "question": "what is the capital of Italy?", "answer": "Rome"}

    Args:
        manifest_filepath: Path to manifest json as described above. Can be comma-separated paths.
        text_processor: TextProcessing object
        sample_rate (int): Sample rate to resample loaded audio to
        int_values (bool): If true, load samples as 32-bit integers. Defauts to False.
        augmentor (nemo.collections.asr.parts.perturb.AudioAugmentor): AudioAugmentor to augment loaded audio
        max_duration: If audio exceeds this length, do not include in dataset
        min_duration: If audio is less than this length, do not include in dataset
        max_utts: Limit number of utterances
        trim: whether or not to trim silence. Defaults to False
        channel_selector (int | Iterable[int] | str): select a single channel or a subset of channels from
            multi-channel audio. If set to `'average'`, it performs averaging across channels.
            Disabled if set to `None`. Defaults to `None`. Uses zero-based indexing.

            :note: below args are NLP-specific

        max_seq_length (int): maximum sequence length for each dataset examples. Examples will either be truncated
            to fit this length or dropped if they cannot be truncated.
        min_seq_length (int): min length of each data example in the dataset. Data examples will be dropped if they
            do not meet the min length requirements.
        tokens_to_generate (int): max num of tokens to generate in a single pass. Defaults to 128.
        pad_to_max_length (bool): If True, pad all samples to the max_seq_length. If False, pad to the max length
        context_key: Key to use for the context in your JSONL file
        answer_key: Key to use for the label in your JSONL file
        context_file: Optional[Union[List[str], str]] = None, if provided, will use this file to load random
            questions from, if question is not in manifest.
    """

    def __init__(
        self,
        manifest_filepath: str,
        text_processor: TextProcessing,
        sample_rate: int,
        int_values: bool = False,
        augmentor: 'nemo.collections.asr.parts.perturb.AudioAugmentor' = None,
        max_duration: Optional[int] = None,
        min_duration: Optional[int] = None,
        max_utts: int = 0,
        trim: bool = False,
        channel_selector: Optional[ChannelSelectorType] = None,
        max_seq_length: int = 1024,
        min_seq_length: int = 1,
        tokens_to_generate: int = 128,
        pad_to_max_length: bool = False,
        max_num_samples: Optional[int] = None,
        index_by_file_id: bool = False,
        context_key: str = 'context',
        answer_key: str = 'answer',
        context_file: Optional[Union[List[str], str]] = None,
    ):
        super().__init__()
        self.text_processor = text_processor
        self.max_seq_length = max_seq_length
        self.min_seq_length = min_seq_length
        self.tokens_to_generate = tokens_to_generate
        self.pad_to_max_length = pad_to_max_length

        if isinstance(manifest_filepath, str):
            manifest_filepath = manifest_filepath.split(",")

        # If necessary, cache manifests and audio from object store
        cache_datastore_manifests(manifest_filepaths=manifest_filepath, cache_audio=True)

        self.collection = collections.SpeechLLMAudioTextCollection(
            manifests_files=manifest_filepath,
            min_duration=min_duration,
            max_duration=max_duration,
            max_number=max_utts,
            index_by_file_id=index_by_file_id,
            max_num_samples=max_num_samples,
            context_file=context_file,
            context_key=context_key,
            answer_key=answer_key,
        )

        self.featurizer = WaveformFeaturizer(sample_rate=sample_rate, int_values=int_values, augmentor=augmentor)
        self.trim = trim
        self.channel_selector = channel_selector

    def get_manifest_sample(self, sample_id):
        """
        return the manifest item of the given index
        """
        return self.collection[sample_id]

    def __getitem__(self, index):
        output = {"idx": index}
        sample = self.collection[index]
        offset = sample.offset

        if offset is None:
            offset = 0

        if sample.audio_file is not None:
            features = self.featurizer.process(
                sample.audio_file,
                offset=offset,
                duration=sample.duration,
                trim=self.trim,
                orig_sr=sample.orig_sr,
                channel_selector=self.channel_selector,
            )
            f, fl = features, torch.tensor(features.shape[0]).long()
            output["audio_signal"] = f
            output["audio_length"] = fl
        else:
            # dummy features
            output["audio_signal"] = torch.zeros([80])
            # accomodates normalize_batch
            output["audio_length"] = torch.tensor(80)

        text_data = self.text_processor(context=sample.context, output=sample.answer)

        output.update(text_data)
        output['metadata'] = {
            'audio_filepath': sample.audio_file,
            'offset': offset,
            'duration': sample.duration,
        }
        return output

    def __len__(self):
        return len(self.collection)

    def _collate_fn(self, batch):
        return _speechllm_audio_text_collate_fn(
            batch=batch,
            tokens_to_generate=self.tokens_to_generate,
            pad_to_max_length=self.pad_to_max_length,
            max_seq_length=self.max_seq_length,
            text_pad_id=self.text_processor.pad_id,
        )

    def collate_fn(self, batch):
        # override collate_fn to skip type checking
        return self._collate_fn(batch)


class MultiAudioTextDataset(AudioTextDataset):
    """
    Dataset for having multi audios per sample, for example in few-shot in-context learning.
    To use this dataset, you need to specify the `audio_locator` field in the dataset config,
    and use that to specify the locations of the audio files in your manifest. In this case,
    the `audio_filepath` field in the manifest is a list of audio filepaths, and the `duration`
    field is a list of durations, one for each audio file. The `offset` field is optional, and
    if not specified, it is assumed to be 0.0. The `offset` field is also a list of offsets if specified.

    Example manifest item for audio_locator='|audio|':
    {
    "audio_filepath": ["1.wav","2.wav","3.wav"],
    "duration": [1.05,1.05,2.0],
    "answer": "this was her dream as nearly as she could recall it",
    "question": "Following are examples of speech audios and their transcriptions.
        Example 1: audio is |audio|, transcription is 'I have a dream'.
        Example 2: audio is |audio|, transcription is ' I don't have a dream'.
        Given the following audio |audio|, transcribe the audio into words."
    }
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    def _collate_fn(self, batch):
        return _speechllm_multi_audio_text_collate_fn(
            batch=batch,
            tokens_to_generate=self.tokens_to_generate,
            pad_to_max_length=self.pad_to_max_length,
            max_seq_length=self.max_seq_length,
            text_pad_id=self.text_processor.pad_id,
        )

    def __getitem__(self, index):
        output = {"idx": index}
        sample = self.collection[index]
        offsets = sample.offset if sample.offset else 0.0
        durations = sample.duration if sample.duration else 0.0
        num_audios = 0
        output["audio_signal"] = []
        output["audio_length"] = []
        if sample.audio_file is not None:
            audio_list = sample.audio_file
            if isinstance(sample.audio_file, str):
                audio_list = [sample.audio_file]
            if not isinstance(audio_list, list):
                raise ValueError(
                    f"The field `audio_file` must be either a str or a list of str,",
                    f"but got type {type(sample.audio_file)} instead",
                )

            num_audios = len(audio_list)
            if isinstance(durations, list) and len(durations) != num_audios:
                raise ValueError(
                    f"The number of durations ({len(durations)}) must match the number of audio clips ({num_audios})"
                )
            if isinstance(offsets, list) and len(offsets) != num_audios:
                raise ValueError(
                    f"The number of offsets ({len(offsets)}) must match the number of audio clips ({num_audios})"
                )

            for i, audio_file in enumerate(audio_list):
                duration = durations[i] if isinstance(durations, list) else 0
                offset = offsets[i] if isinstance(offsets, list) else 0
                features = self.featurizer.process(
                    audio_file,
                    offset=offset,
                    duration=duration,
                    trim=self.trim,
                    orig_sr=sample.orig_sr,
                    channel_selector=self.channel_selector,
                )
                f, fl = features, torch.tensor(features.shape[0]).long()
                output["audio_signal"].append(f)
                output["audio_length"].append(fl)
        else:
            # dummy features
            output["audio_signal"] = [torch.zeros([8])]
            # accomodates normalize_batch
            output["audio_length"] = [torch.tensor(8)]

        text_data = self.text_processor(context=sample.context, output=sample.answer)

        if isinstance(output["audio_signal"], list) and len(output["audio_signal"]) + 1 != len(
            text_data['context_start_idx']
        ):
            raise ValueError(
                f"The number of text segments ({len(text_data['context_start_idx'])})",
                f"must be one more than number of audios ({len(output['audio_signal'])})",
            )

        output.update(text_data)
        output['metadata'] = {
            'audio_filepath': sample.audio_file,
            'offset': offsets,
            'duration': sample.duration,
        }
        return output


class TarredAudioFilter:
    """
    filter function for tarred audio files, skip entry if not in manifest
    """

    def __init__(self, collection, iterator):
        self.iterator = iterator
        self.collection = collection

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            audio_bytes, audio_filename = next(self.iterator)
            file_id, _ = os.path.splitext(os.path.basename(audio_filename))
            if file_id in self.collection.mapping:
                return audio_bytes, audio_filename
            else:
                logging.warning(f"key not in manifest: {file_id}", mode=logging_mode.ONCE)


class TarredAudioLoopOffsets:
    """
    Loop over tarred audio files
    """

    def __init__(self, collection, iterator):
        self.iterator = iterator
        self.collection = collection
        self.current_fn = None
        self.current_bytes = None
        self.offset_id = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_fn is None:
            self.current_bytes, self.current_fn = next(self.iterator)
            self.offset_id = 0
        else:
            offset_list = self.collection.mapping[self.current_fn]
            if len(offset_list) == self.offset_id + 1:
                self.current_bytes, self.current_fn = next(self.iterator)
                self.offset_id = 0
            else:
                self.offset_id += 1

        return self.current_bytes, self.current_fn, self.offset_id


class TarredAudioTextDataset(IterableDataset):
    """
    A similar Dataset to the AudioTextDataset, but which loads tarred audio files.

    Accepts a single comma-separated JSON manifest file (in the same style as for the AudioTextDataset),
    as well as the path(s) to the tarball(s) containing the wav files. Each line of the manifest should
    contain the information for one audio file, including at least the transcript and name of the audio
    file within the tarball.

    Valid formats for the audio_tar_filepaths argument include:
    (1) a single string that can be brace-expanded, e.g. 'path/to/audio.tar' or 'path/to/audio_{1..100}.tar.gz', or
    (2) a list of file paths that will not be brace-expanded, e.g. ['audio_1.tar', 'audio_2.tar', ...].

    Note: For brace expansion in (1), there may be cases where `{x..y}` syntax can't be used due to shell.
    This occurs most commonly inside SLURM scripts. Therefore we provide a few equivalent replacements.
    Supported opening braces - { <=> (, [, < and the special tag _OP_.
    Supported closing braces - } <=> ), ], > and the special tag _CL_.
    For SLURM based tasks, we suggest the use of the special tags for ease of use.

    See the WebDataset documentation for more information about accepted data and input formats.

    If using multiple workers the number of shards should be divisible by world_size to ensure an
    even split among workers. If it is not divisible, logging will give a warning but training will proceed.
    In addition, if using mutiprocessing, each shard MUST HAVE THE SAME NUMBER OF ENTRIES after filtering
    is applied. We currently do not check for this, but your program may hang if the shards are uneven!

    Additionally, please note that the len() of this DataLayer is assumed to be the length of the manifest
    after filtering. An incorrect manifest length may lead to some DataLoader issues down the line.

    Args:
        audio_tar_filepaths: Either a list of audio tarball filepaths, or a
            string (can be brace-expandable).
        manifest_filepath (str): Path to the manifest.
        text_processor: TextProcessing object,
        sample_rate (int): Sample rate to resample loaded audio to
        int_values (bool): If true, load samples as 32-bit integers. Defauts to False.
        augmentor (nemo.collections.asr.parts.perturb.AudioAugmentor): An AudioAugmentor
            object used to augment loaded audio
        shuffle_n (int): How many samples to look ahead and load to be shuffled.
            See WebDataset documentation for more details.
            Defaults to 0.
        min_duration (float): Dataset parameter.
            All training files which have a duration less than min_duration
            are dropped. Note: Duration is read from the manifest JSON.
            Defaults to 0.1.
        max_duration (float): Dataset parameter.
            All training files which have a duration more than max_duration
            are dropped. Note: Duration is read from the manifest JSON.
            Defaults to None.
        blank_index (int): Blank character index, defaults to -1.
        unk_index (int): Unknown character index, defaults to -1.
        normalize (bool): Dataset parameter.
            Whether to use automatic text cleaning.
            It is highly recommended to manually clean text for best results.
            Defaults to True.
        trim (bool): Whether to use trim silence from beginning and end
            of audio signal using librosa.effects.trim().
            Defaults to False.
        bos_id (id): Dataset parameter.
            Beginning of string symbol id used for seq2seq models.
            Defaults to None.
        eos_id (id): Dataset parameter.
            End of string symbol id used for seq2seq models.
            Defaults to None.
        pad_id (id): Token used to pad when collating samples in batches.
            If this is None, pads using 0s.
            Defaults to None.
        shard_strategy (str): Tarred dataset shard distribution strategy chosen as a
            str value during ddp.

            - `scatter`: The default shard strategy applied by WebDataset, where each node gets
              a unique set of shards, which are permanently pre-allocated and never changed at runtime.
            - `replicate`: Optional shard strategy, where each node gets all of the set of shards
              available in the tarred dataset, which are permanently pre-allocated and never changed at runtime.
              The benefit of replication is that it allows each node to sample data points from the entire
              dataset independently of other nodes, and reduces dependence on value of `shuffle_n`.

            :warning: Replicated strategy allows every node to sample the entire set of available tarfiles,
                and therefore more than one node may sample the same tarfile, and even sample the same
                data points! As such, there is no assured guarantee that all samples in the dataset will be
                sampled at least once during 1 epoch. Scattered strategy, on the other hand, on specific
                occasions (when the number of shards is not divisible with ``world_size``), will not sample
                the entire dataset. For these reasons it is not advisable to use tarred datasets as validation
                or test datasets.

        shard_manifests (bool): Whether or not to try / shard manifests. Defaults to False.
        global_rank (int): Worker rank, used for partitioning shards. Defaults to 0.
        world_size (int): Total number of processes, used for partitioning shards. Defaults to 0.

            :note: Below args are NLP-specific

        max_seq_length (int): maximum sequence length for each dataset examples. Examples will either be truncated
            to fit this length or dropped if they cannot be truncated.
        min_seq_length (int): min length of each data example in the dataset. Data examples will be dropped if they
            do not meet the min length requirements.
        tokens_to_generate (int): maximum tokens to generate in a single pass. Defaults to 128.
        context_key: Key to use for the context in your JSONL file
        answer_key: Key to use for the label in your JSONL file
        context_file: Optional[Union[List[str], str]] = None, if provided, will use this file to load
            random questions from, if question is not in manifest.
    """

    def __init__(
        self,
        audio_tar_filepaths: Union[str, List[str]],
        manifest_filepath: str,
        text_processor: TextProcessing,
        sample_rate: int,
        int_values: bool = False,
        augmentor: Optional['nemo.collections.asr.parts.perturb.AudioAugmentor'] = None,
        shuffle_n: int = 0,
        min_duration: Optional[float] = None,
        max_duration: Optional[float] = None,
        trim: bool = False,
        shard_strategy: str = "scatter",
        shard_manifests: bool = False,
        global_rank: int = 0,
        world_size: int = 0,
        max_seq_length: int = 1024,
        min_seq_length: int = 1,
        tokens_to_generate: int = 128,
        pad_to_max_length: bool = False,
        context_key: str = 'context',
        answer_key: str = 'answer',
        context_file: Optional[Union[List[str], str]] = None,
    ):
        super().__init__()
        self.text_processor = text_processor
        self.max_seq_length = max_seq_length
        self.min_seq_length = min_seq_length
        self.is_megatron_iterable = True
        self.shard_manifests = shard_manifests
        self.tokens_to_generate = tokens_to_generate
        self.pad_to_max_length = pad_to_max_length

        # Shard manifests if necessary and possible and then expand the paths
        manifest_filepath = shard_manifests_if_needed(
            shard_manifests=shard_manifests,
            shard_strategy=shard_strategy,
            manifest_filepaths=manifest_filepath,
            world_size=world_size,
            global_rank=global_rank,
        )

        # If necessary, cache manifests from object store
        cache_datastore_manifests(manifest_filepaths=manifest_filepath)

        self.collection = collections.SpeechLLMAudioTextCollection(
            manifests_files=manifest_filepath,
            min_duration=min_duration,
            max_duration=max_duration,
            index_by_file_id=True,
            context_file=context_file,
            context_key=context_key,
            answer_key=answer_key,
        )

        self.len = self._compute_len()

        self.featurizer = WaveformFeaturizer(sample_rate=sample_rate, int_values=int_values, augmentor=augmentor)
        self.trim = trim

        audio_tar_filepaths = expand_sharded_filepaths(
            sharded_filepaths=audio_tar_filepaths,
            shard_strategy=shard_strategy,
            world_size=world_size,
            global_rank=global_rank,
        )

        # Put together WebDataset
        self._dataset = wds.WebDataset(urls=audio_tar_filepaths, nodesplitter=None)

        if shuffle_n == 0:
            logging.info("WebDataset will not shuffle files within the tar files.")

        # Put together WebDataset pipeline
        self._dataset = wds.DataPipeline(
            wds.SimpleShardList(urls=audio_tar_filepaths),
            webdataset_split_by_workers,
            wds.shuffle(shuffle_n),
            wds.tarfile_to_samples(),
            wds.rename(audio=VALID_FILE_FORMATS, key='__key__'),
            wds.to_tuple('audio', 'key'),
            self._filter,
            self._loop_offsets,
            wds.map(self._build_sample),
        )

    def _filter(self, iterator):
        """This function is used to remove samples that have been filtered out by ASRAudioText already.
        Otherwise, we would get a KeyError as _build_sample attempts to find the manifest entry for a sample
        that was filtered out (e.g. for duration).
        Note that if using multi-GPU training, filtering may lead to an imbalance in samples in each shard,
        which may make your code hang as one process will finish before the other.
        """
        return TarredAudioFilter(self.collection, iterator)

    def _loop_offsets(self, iterator):
        """This function is used to iterate through utterances with different offsets for each file."""
        return TarredAudioLoopOffsets(self.collection, iterator)

    def _collate_fn(self, batch):
        return _speechllm_audio_text_collate_fn(
            batch=batch,
            tokens_to_generate=self.tokens_to_generate,
            pad_to_max_length=self.pad_to_max_length,
            max_seq_length=self.max_seq_length,
            text_pad_id=self.text_processor.pad_id,
        )

    def collate_fn(self, batch):
        # override collate_fn to skip type checking
        return self._collate_fn(batch)

    def _build_sample(self, tup):
        """Builds the training sample by combining the data from the WebDataset with the manifest info."""
        audio_bytes, audio_filename, offset_id = tup

        if audio_filename is not None:
            # Grab manifest entry from self.manifest_preprocessor.collection
            file_id, _ = os.path.splitext(os.path.basename(audio_filename))
            manifest_idx = self.collection.mapping[file_id][offset_id]
            manifest_entry = self.collection[manifest_idx]

            # init output dict
            output = {"idx": manifest_idx}

            offset = manifest_entry.offset
            if offset is None:
                offset = 0
            # Convert audio bytes to IO stream for processing (for SoundFile to read)
            audio_filestream = io.BytesIO(audio_bytes)
            features = self.featurizer.process(
                audio_filestream,
                offset=offset,
                duration=manifest_entry.duration,
                trim=self.trim,
                orig_sr=manifest_entry.orig_sr,
            )
            audio_filestream.close()

            # Audio features
            output["audio_signal"] = features
            output["audio_length"] = torch.tensor(features.shape[0]).long()
        else:
            # dummy features
            output["audio_signal"] = torch.zeros([80])
            # accomodates normalize_batch
            output["audio_length"] = torch.tensor(80)

        # Text features
        text_data = self.text_processor(context=manifest_entry.context, output=manifest_entry.answer)

        output.update(text_data)

        output['metadata'] = {
            'audio_filepath': audio_filename,
            'offset': offset,
            'duration': manifest_entry.duration,
        }
        return output

    def get_manifest_sample(self, sample_id):
        """
        return manifest item given the index
        """
        return self.collection[sample_id]

    def __iter__(self):
        return self._dataset.__iter__()

    def _compute_len(self):
        # TODO: need to figure out why here needs to be divided by world_size, while in ASR we don't need to.
        if self.shard_manifests and torch.distributed.is_available() and torch.distributed.is_initialized():
            my_len = torch.tensor(len(self.collection), dtype=torch.int32).cuda()
            torch.distributed.all_reduce(my_len)
            my_len = my_len.int() // parallel_state.get_data_parallel_world_size()
            logging.info(f'Sharded manifests: Total length: {my_len}')
        else:
            my_len = len(self.collection) // parallel_state.get_data_parallel_world_size()

        return my_len

    def __len__(self):
        return self.len


def get_tarred_audio_text_dataset(
    config,
    text_processor,
    augmentor,
    global_rank=0,
    world_size=1,
    shuffle_n=0,
):
    """
    Get tarred audio to text dataset
    """
    tarred_audio_filepaths = config['tarred_audio_filepaths']
    manifest_filepaths = config['manifest_filepath']
    datasets = []
    tarred_audio_filepaths = convert_to_config_list(tarred_audio_filepaths)
    manifest_filepaths = convert_to_config_list(manifest_filepaths)

    bucketing_weights = config.get('bucketing_weights', None)  # For upsampling buckets
    if bucketing_weights:
        for idx, weight in enumerate(bucketing_weights):
            if not isinstance(weight, int) or weight <= 0:
                raise ValueError(f"bucket weights must be positive integers")

    if len(manifest_filepaths) != len(tarred_audio_filepaths):
        raise ValueError(
            f"manifest_filepaths (length={len(manifest_filepaths)}) and tarred_audio_filepaths",
            f"(length={len(tarred_audio_filepaths)}) need to have the same number of buckets.",
        )

    if 'labels' not in config:
        logging.warning(f"dataset does not have explicitly defined labels")

    if 'max_utts' in config:
        raise ValueError('"max_utts" parameter is not supported for tarred datasets')

    for dataset_idx, (tarred_audio_filepath, manifest_filepath) in enumerate(
        zip(tarred_audio_filepaths, manifest_filepaths)
    ):
        if len(tarred_audio_filepath) == 1:
            tarred_audio_filepath = tarred_audio_filepath[0]
        if len(manifest_filepath) == 1:
            manifest_filepath = manifest_filepath[0]

        dataset = TarredAudioTextDataset(
            audio_tar_filepaths=tarred_audio_filepath,
            manifest_filepath=manifest_filepath,
            text_processor=text_processor,
            sample_rate=config['sample_rate'],
            int_values=config.get('int_values', False),
            augmentor=augmentor,
            shuffle_n=shuffle_n,
            max_duration=config.get('max_duration', None),
            min_duration=config.get('min_duration', None),
            trim=config.get('trim_silence', False),
            shard_strategy=config.get('tarred_shard_strategy', 'scatter'),
            shard_manifests=config.get('shard_manifests', False),
            global_rank=global_rank,
            world_size=world_size,
            max_seq_length=config.max_seq_length,
            min_seq_length=config.min_seq_length,
            tokens_to_generate=config.get('tokens_to_generate', 0),
            pad_to_max_length=config.get('pad_to_max_length', False),
            context_key=config.get('context_key', 'context'),
            answer_key=config.get('answer_key', 'answer'),
            context_file=config.get('context_file', None),
        )

        if bucketing_weights:
            [datasets.append(dataset) for _ in range(bucketing_weights[dataset_idx])]
        else:
            datasets.append(dataset)

    with open_dict(config):  # patch for bucketing tarred datasets
        config['batch_size'] = config.get("micro_batch_size", 1)
    return get_chain_dataset(datasets=datasets, ds_config=config, rank=global_rank)


def get_concat_tarred_audio_text_dataset(
    config,
    text_processor,
    augmentor,
    global_rank=0,
    world_size=1,
    shuffle_n=0,
):
    """
    Get concat tarred audio to text dataset
    """
    tarred_audio_filepaths = config['tarred_audio_filepaths']
    manifest_filepaths = config['manifest_filepath']
    datasets = []
    for dataset_idx, (tarred_audio_filepath, manifest_filepath) in enumerate(
        zip(tarred_audio_filepaths, manifest_filepaths)
    ):
        conf = copy.deepcopy(config)
        conf['manifest_filepath'] = manifest_filepath
        conf['tarred_audio_filepaths'] = tarred_audio_filepath
        context_files = config.get('context_file', None)
        if isinstance(context_files, ListConfig) and len(context_files) == len(manifest_filepaths):
            conf['context_file'] = context_files[dataset_idx]
        else:
            conf['context_file'] = context_files
        dataset = get_tarred_audio_text_dataset(
            config=conf,
            text_processor=text_processor,
            shuffle_n=shuffle_n,
            global_rank=global_rank,
            world_size=world_size,
            augmentor=augmentor,
        )
        datasets.append(dataset)

    concat_sampling_probabilities = config.get('concat_sampling_probabilities', None)
    if not isinstance(concat_sampling_probabilities, ListConfig) or len(concat_sampling_probabilities) != len(
        datasets
    ):
        logging.info(
            f"concat_sampling_probabilities is not provided or is not of the same size as datasets,"
            f"using uniform sampling: concat_sampling_probabilities={concat_sampling_probabilities}"
        )
        concat_sampling_probabilities = [1.0 / len(datasets)] * len(datasets)

    dataset = ConcatDataset(
        datasets,
        sampling_technique=config.get('concat_sampling_technique', 'temperature'),
        sampling_temperature=config.get('concat_sampling_temperature', 5),
        sampling_scale=config.get('concat_sampling_scale', 1),
        sampling_probabilities=concat_sampling_probabilities,
        shuffle=config.get('concat_shuffle', True),
        seed=config.get('concat_sampling_seed', None),
        global_rank=global_rank,
        world_size=world_size,
    )
    return dataset


def get_tarred_audio_text_dataset_from_config(
    config: DictConfig,
    text_processor: TextProcessing,
    augmentor,
    global_rank: int = 0,
    world_size: int = 1,
):
    """
    Get tarred dataset from config
    """
    is_concat = config.get('is_concat', False)
    if is_concat:
        if 'concat_sampling_technique' in config and config['concat_sampling_technique'] is None:
            logging.warning(
                f"Concat dataset requires `concat_sampling_technique` but it was not provided. Config: {config}"
            )
            return None

    data_parallel_size = parallel_state.get_data_parallel_world_size()
    num_micro_batches = config.global_batch_size // (config.micro_batch_size * data_parallel_size)
    global_batch_size_on_this_data_parallel_rank = num_micro_batches * config.micro_batch_size
    shuffle = config['shuffle']
    shuffle_n = config.get('shuffle_n', 4 * global_batch_size_on_this_data_parallel_rank) if shuffle else 0
    if is_concat:
        dataset = get_concat_tarred_audio_text_dataset(
            config=config,
            text_processor=text_processor,
            augmentor=augmentor,
            shuffle_n=shuffle_n,
            global_rank=global_rank,
            world_size=world_size,
        )
    else:
        dataset = get_tarred_audio_text_dataset(
            config=config,
            text_processor=text_processor,
            augmentor=augmentor,
            shuffle_n=shuffle_n,
            global_rank=global_rank,
            world_size=world_size,
        )
    return dataset


def get_audio_text_dataset_from_config(
    manifest_filepath: str,
    config: DictConfig,
    text_processor: TextProcessing,
    augmentor,
    is_train,
):
    """
    Get non-tarred aduio to text dataset from config
    """
    if isinstance(config.manifest_filepath, str):
        manifest_filepath = config.manifest_filepath.split(',')
    else:
        manifest_filepath = config.manifest_filepath

    data_cls = MultiAudioTextDataset if config.get('audio_locator', None) else AudioTextDataset
    logging.info(f"Using `{data_cls.__name__}` for dataset")
    datasets = []

    if is_train:
        # Construct the data prefix list for `get_datasets_weights_and_num_samples()`
        # that is of the format [weight1,file_name1,weight2,file_name2,...]
        concat_sampling_probabilities = config.get('concat_sampling_probabilities', None)
        if concat_sampling_probabilities is None:
            concat_sampling_probabilities = [1.0 / len(manifest_filepath)] * len(manifest_filepath)
        elif len(config.get('concat_sampling_probabilities', None)) != len(manifest_filepath):
            raise ValueError(
                (
                    f"concat_sampling_probabilities must be of the same size as manifest_filepath.",
                    f"Provided size {len(config.concat_sampling_probabilities)},",
                    f"number of datasets {len(manifest_filepath)}",
                )
            )
        data_prefix = []
        for weight, prefix in zip(concat_sampling_probabilities, manifest_filepath):
            data_prefix.append(weight)
            data_prefix.append(prefix)

        num_samples_per_dataset = get_num_samples_from_files(manifest_filepath)
        num_train_samples = [len(manifest_filepath) * max(num_samples_per_dataset)]
        _, _, num_train_samples_per_dataset = get_datasets_weights_and_num_samples(data_prefix, num_train_samples)
        num_train_samples_after_blend = sum([x[0] for x in num_train_samples_per_dataset])

    # for dataset_idx, (file_path, num_samples) in enumerate(zip(manifest_filepath, num_train_samples_per_dataset)):
    for dataset_idx, file_path in enumerate(manifest_filepath):
        context_file = config.get('context_file', None)
        if isinstance(context_file, ListConfig) and len(context_file) == len(manifest_filepath):
            context_file = context_file[dataset_idx]
        dataset = data_cls(
            manifest_filepath=file_path,
            text_processor=text_processor,
            sample_rate=config.sample_rate,
            int_values=config.get('int_values', False),
            augmentor=augmentor,
            max_duration=getattr(config, 'max_duration', None),
            min_duration=getattr(config, 'min_duration', None),
            max_utts=getattr(config, 'max_utts', -1),
            trim=getattr(config, 'trim_silence', False),
            channel_selector=getattr(config, 'channel_selector', None),
            max_seq_length=config.max_seq_length,
            min_seq_length=config.min_seq_length,
            tokens_to_generate=config.get('tokens_to_generate', 0),
            pad_to_max_length=config.get('pad_to_max_length', False),
            context_key=config.get('context_key', 'context'),
            answer_key=config.get('answer_key', 'answer'),
            context_file=context_file,
        )
        datasets.append(dataset)

    if is_train:
        dataset = BlendableDataset(
            datasets=datasets, weights=concat_sampling_probabilities, size=num_train_samples_after_blend
        )
        return dataset
    else:
        return datasets


def get_datasets_weights_and_num_samples(data_prefix, num_samples):

    # The data prefix should be in the format of:
    #   weight-1, data-prefix-1, weight-2, data-prefix-2, ..
    assert len(data_prefix) % 2 == 0
    num_datasets = len(data_prefix) // 2
    weights = [0] * num_datasets
    prefixes = [0] * num_datasets
    for i in range(num_datasets):
        weights[i] = float(data_prefix[2 * i])
        prefixes[i] = (data_prefix[2 * i + 1]).strip()
    # Normalize weights
    weight_sum = 0.0
    for weight in weights:
        weight_sum += weight
    assert weight_sum > 0.0
    weights = [weight / weight_sum for weight in weights]

    # Add 0.5% (the 1.005 factor) so in case the bleding dataset does
    # not uniformly distribute the number of samples, we still have
    # samples left to feed to the network.
    # TODO: check data leakage between train/val/test?
    datasets_train_valid_test_num_samples = []
    for weight in weights:
        # Comes here when we have seperate train,test and validation datasets.
        if isinstance(num_samples, int):
            datasets_train_valid_test_num_samples.append(int(math.ceil(num_samples * weight * 1.005)))
        else:
            datasets_train_valid_test_num_samples.append([int(math.ceil(val * weight * 1.005)) for val in num_samples])

    return prefixes, weights, datasets_train_valid_test_num_samples



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import einops
import torch

from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import LengthsType, NeuralType, SpectrogramType

__all__ = ['SSLPretrainWithMaskedPatch']


class SSLPretrainWithMaskedPatch(NeuralModule):
    """
    Zeroes out fixed size time patches of the spectrogram.
    All samples in batch are guaranteed to have the same amount of masked time steps.
    Note that this may be problematic when we do pretraining on a unbalanced dataset.

    For example, say a batch contains two spectrograms of length 87 and 276.
    With mask_fraction=0.7 and patch_size=10, we'll obrain mask_patches=7.
    Each of the two data will then have 7 patches of 10-frame mask.

    Args:
        patch_size (int): up to how many time steps does one patch consist of.
            Defaults to 10.
        mask_fraction (float): how much fraction in each sample to be masked (number of patches is rounded up).
            Range from 0.0 to 1.0. Defaults to 0.7.
    """

    @property
    def input_types(self):
        """Returns definitions of module input types"""
        return {
            "input_spec": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        """Returns definitions of module output types"""
        return {"augmented_spec": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType())}

    def __init__(
        self,
        patch_size: int = 10,
        mask_fraction: float = 0.7,
    ):
        super().__init__()
        if patch_size <= 0:
            raise ValueError(f'patch_size must be positive, got patch_size={patch_size}')

        if mask_fraction > 1.0 or mask_fraction < 0.0:
            raise ValueError(f'mask_fraction must be in the range [0.0, 1.0], got mask_fraction={mask_fraction}')

        self.patch_size = patch_size
        self.mask_fraction = mask_fraction

    @typecheck()
    def forward(self, input_spec, length):
        """
        Apply Patched masking on the input_spec.


        During the training stage, the mask is generated randomly, with
        approximately `self.mask_fraction` of the time frames being masked out.

        In the validation stage, the masking pattern is fixed to ensure
        consistent evaluation of checkpoints and to prevent overfitting. Note
        that the same masking pattern is applied to all data, regardless of
        their lengths. On average, approximately `self.mask_fraction` of the
        time frames will be masked out.

        """
        augmented_spec = input_spec

        if self.training:
            for idx, cur_len in enumerate(length.tolist()):
                # patch indices
                patches = range(cur_len // self.patch_size)
                # fraction of samples to mask
                len_fraction = int(cur_len * self.mask_fraction)
                # number of patches to mask
                mask_patches = len_fraction // self.patch_size + int(len_fraction % self.patch_size != 0)
                # if the number of patches to mask is greater than the number of patches, reduce the number of patches to mask
                if cur_len < self.patch_size * mask_patches:
                    mask_patches = cur_len // self.patch_size
                # sample random patches to mask
                masked_patches = random.sample(patches, mask_patches)
                # mask the sampled patches
                for mp in masked_patches:
                    augmented_spec[idx, :, :, mp * self.patch_size : (mp + 1) * self.patch_size] = 0.0
        else:
            chunk_length = self.patch_size // self.mask_fraction
            mask = torch.arange(augmented_spec.size(-1), device=augmented_spec.device)
            mask = (mask % chunk_length) >= self.patch_size
            mask = einops.rearrange(mask, 'T -> 1 1 1 T').float()
            augmented_spec = augmented_spec * mask

        return augmented_spec



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import inspect
import uuid
from collections.abc import AsyncGenerator
from datetime import datetime
from typing import Iterator, List, Optional

import numpy as np
import torch
from loguru import logger
from pipecat.frames.frames import (
    CancelFrame,
    EndFrame,
    ErrorFrame,
    Frame,
    LLMTextFrame,
    StartFrame,
    TTSAudioRawFrame,
    TTSStartedFrame,
    TTSStoppedFrame,
)
from pipecat.services.llm_service import FunctionCallParams
from pipecat.services.tts_service import TTSService

from nemo.agents.voice_agent.pipecat.services.nemo.audio_logger import AudioLogger
from nemo.agents.voice_agent.pipecat.utils.tool_calling.mixins import ToolCallingMixin
from nemo.collections.tts.models import FastPitchModel, HifiGanModel


class BaseNemoTTSService(TTSService):
    """Text-to-Speech service using Nemo TTS models.

    This service works with any TTS model that exposes a generate(text) method
    that returns audio data. The TTS generation runs in a dedicated background thread to
    avoid blocking the main asyncio event loop, following the same pattern as NemoDiarService.

    Args:
        model: TTS model instance with a generate(text) method
        sample_rate: Audio sample rate in Hz (defaults to 22050)
        **kwargs: Additional arguments passed to TTSService
    """

    def __init__(
        self,
        *,
        model,
        device: str = "cuda",
        sample_rate: int = 22050,
        think_tokens: Optional[List[str]] = None,
        audio_logger: Optional[AudioLogger] = None,
        **kwargs,
    ):
        super().__init__(sample_rate=sample_rate, **kwargs)
        self._model_name = model
        self._device = device
        self._model = self._setup_model()
        self._think_tokens = think_tokens
        self._audio_logger = audio_logger
        if think_tokens is not None:
            assert (
                isinstance(think_tokens, list) and len(think_tokens) == 2
            ), f"think_tokens must be a list of two strings: {think_tokens}"

        # Background processing infrastructure - no response handler needed
        self._tts_queue = asyncio.Queue()
        self._processing_task = None
        self._processing_running = False

        # Track pending requests with their response queues
        self._pending_requests = {}
        self._have_seen_think_tokens = False

    def _setup_model(self):
        raise NotImplementedError("Subclass must implement _setup_model")

    def _generate_audio(self, text: str) -> Iterator[np.ndarray]:
        raise NotImplementedError("Subclass must implement _generate_audio")

    def can_generate_metrics(self) -> bool:
        """If the TTS service can generate metrics."""
        return True

    async def start(self, frame: StartFrame):
        """Handle service start."""
        await super().start(frame)

        # Initialize the model if not already done
        if not hasattr(self, "_model") or self._model is None:
            self._model = self._setup_model()

        # Only start background processing task - no response handler needed
        if not self._processing_task:
            self._processing_task = self.create_task(self._processing_task_handler())

    async def stop(self, frame: EndFrame):
        """Handle service stop."""
        await super().stop(frame)
        await self._stop_tasks()

    async def cancel(self, frame: CancelFrame):
        """Handle service cancellation."""
        await super().cancel(frame)
        await self._stop_tasks()

    async def _stop_tasks(self):
        """Stop background processing tasks."""
        self._processing_running = False
        await self._tts_queue.put(None)  # Signal to stop processing

        if self._processing_task:
            await self.cancel_task(self._processing_task)
            self._processing_task = None

    def _tts_processor(self):
        """Background processor that handles TTS generation calls."""
        try:
            while self._processing_running:
                try:
                    future = asyncio.run_coroutine_threadsafe(self._tts_queue.get(), self.get_event_loop())
                    request = future.result()

                    if request is None:  # Stop signal
                        logger.debug("Received stop signal in TTS background processor")
                        break

                    text, request_id = request
                    logger.debug(f"Processing TTS request for text: [{text}]")

                    # Get the response queue for this request
                    response_queue = None
                    future = asyncio.run_coroutine_threadsafe(
                        self._get_response_queue(request_id), self.get_event_loop()
                    )
                    response_queue = future.result()

                    if response_queue is None:
                        logger.warning(f"No response queue found for request {request_id}")
                        continue

                    # Process TTS generation
                    try:
                        audio_result = self._generate_audio(text)

                        # Send result directly to the waiting request
                        asyncio.run_coroutine_threadsafe(
                            response_queue.put(('success', audio_result)), self.get_event_loop()
                        )
                    except Exception as e:
                        logger.error(f"Error in TTS generation: {e}")
                        # Send error directly to the waiting request
                        asyncio.run_coroutine_threadsafe(response_queue.put(('error', e)), self.get_event_loop())

                except Exception as e:
                    logger.error(f"Error in background TTS processor: {e}")

        except Exception as e:
            logger.error(f"Background TTS processor fatal error: {e}")
        finally:
            logger.debug("Background TTS processor stopped")

    async def _get_response_queue(self, request_id: str):
        """Get the response queue for a specific request."""
        return self._pending_requests.get(request_id)

    async def _processing_task_handler(self):
        """Handler for background processing task."""
        try:
            self._processing_running = True
            logger.debug("Starting background TTS processing task")
            await asyncio.to_thread(self._tts_processor)
        except asyncio.CancelledError:
            logger.debug("Background TTS processing task cancelled")
            self._processing_running = False
            raise
        finally:
            self._processing_running = False

    def _handle_think_tokens(self, text: str) -> Optional[str]:
        """
        Handle the thinking tokens for TTS.
        If the thinking tokens are not provided, return the text as it is.
        Otherwise:
            If both thinking tokens appear in the text, return the text after the end of thinking tokens.
            If the LLM is thinking, return None.
            If the LLM is done thinking, return the text after the end of thinking tokens.
            If the LLM starts thinking, return the text before the start of thinking tokens.
            If the LLM is not thinking, return the text as is.
        """
        if not self._think_tokens:
            return text
        elif self._think_tokens[0] in text and self._think_tokens[1] in text:
            # LLM finishes thinking in one chunk or outputs dummy thinking tokens
            logger.debug(f"LLM finishes thinking: {text}")
            idx = text.index(self._think_tokens[1])
            # only return the text after the end of thinking tokens
            text = text[idx + len(self._think_tokens[1]) :]
            self._have_seen_think_tokens = False
            logger.debug(f"Returning text after thinking: {text}")
            return text
        elif self._have_seen_think_tokens:
            # LLM is thinking
            if self._think_tokens[1] not in text:
                logger.debug(f"LLM is still thinking: {text}")
                # LLM is still thinking
                return None
            else:
                # LLM is done thinking
                logger.debug(f"LLM is done thinking: {text}")
                idx = text.index(self._think_tokens[1])
                # only return the text after the end of thinking tokens
                text = text[idx + len(self._think_tokens[1]) :]
                self._have_seen_think_tokens = False
                logger.debug(f"Returning text after thinking: {text}")
                return text
        elif self._think_tokens[0] in text:
            # LLM now starts thinking
            logger.debug(f"LLM starts thinking: {text}")
            self._have_seen_think_tokens = True
            # return text before the start of thinking tokens
            idx = text.index(self._think_tokens[0])
            text = text[:idx]
            logger.debug(f"Returning text before thinking: {text}")
            return text
        else:
            # LLM is not thinking
            return text

    async def run_tts(self, text: str) -> AsyncGenerator[Frame, None]:
        """Generate speech from text using the Nemo TTS model."""
        text = self._handle_think_tokens(text)

        if not text:
            yield None
            return

        logger.debug(f"{self}: Generating TTS [{text}]")

        try:
            await self.start_ttfb_metrics()
            yield TTSStartedFrame()

            # Increment turn index at the start of agent speaking (only if speaker changed)
            if self._audio_logger is not None:
                self._audio_logger.increment_turn_index(speaker="agent")

            # Generate unique request ID

            request_id = str(uuid.uuid4())

            # Create response queue for this specific request
            request_queue = asyncio.Queue()
            self._pending_requests[request_id] = request_queue

            try:
                # Queue the TTS request for background processing
                await self._tts_queue.put((text, request_id))

                # Wait for the result directly from our request queue
                result = await request_queue.get()
                status, data = result

                if status == 'error':
                    logger.error(f"{self} TTS generation error: {data}")
                    yield ErrorFrame(error=f"TTS generation error: {str(data)}")
                    return

                audio_result = data
                if audio_result is None:
                    logger.error(f"{self} TTS model returned None for text: [{text}]")
                    yield ErrorFrame(error="TTS generation failed - no audio returned")
                    return

                await self.start_tts_usage_metrics(text)

                # Collect all audio for logging
                all_audio_bytes = b""
                # Capture the start time when TTS begins (not when it ends)
                if self._audio_logger is not None and self._audio_logger.first_audio_timestamp is None:
                    self._audio_logger.first_audio_timestamp = datetime.now()

                # Process the audio result (same as before)
                if (
                    inspect.isgenerator(audio_result)
                    or hasattr(audio_result, '__iter__')
                    and hasattr(audio_result, '__next__')
                ):
                    # Handle generator case
                    first_chunk = True
                    for audio_chunk in audio_result:
                        if first_chunk:
                            await self.stop_ttfb_metrics()
                            first_chunk = False
                            # Capture start time on first chunk
                            if self._audio_logger is not None:
                                tts_start_time = self._audio_logger.get_time_from_start_of_session()

                        if audio_chunk is None:
                            break

                        audio_bytes = self._convert_to_bytes(audio_chunk)
                        all_audio_bytes += audio_bytes
                        chunk_size = self.chunk_size
                        for i in range(0, len(audio_bytes), chunk_size):
                            audio_chunk_bytes = audio_bytes[i : i + chunk_size]
                            if not audio_chunk_bytes:
                                break

                            frame = TTSAudioRawFrame(
                                audio=audio_chunk_bytes, sample_rate=self.sample_rate, num_channels=1
                            )
                            yield frame
                else:
                    # Handle single result case
                    await self.stop_ttfb_metrics()
                    # Capture start time for single result
                    if self._audio_logger is not None:
                        tts_start_time = self._audio_logger.get_time_from_start_of_session()
                    audio_bytes = self._convert_to_bytes(audio_result)
                    all_audio_bytes = audio_bytes

                    chunk_size = self.chunk_size
                    for i in range(0, len(audio_bytes), chunk_size):
                        chunk = audio_bytes[i : i + chunk_size]
                        if not chunk:
                            break

                        frame = TTSAudioRawFrame(audio=chunk, sample_rate=self.sample_rate, num_channels=1)
                        yield frame

                # Log the complete audio if logger is available
                if self._audio_logger is not None and all_audio_bytes:
                    try:
                        self._audio_logger.log_agent_audio(
                            audio_data=all_audio_bytes,
                            text=text,
                            sample_rate=self.sample_rate,
                            num_channels=1,
                            additional_metadata={
                                "model": self._model_name,
                            },
                            tts_generation_time=tts_start_time,
                        )
                    except Exception as e:
                        logger.warning(f"Failed to log agent audio: {e}")

                yield TTSStoppedFrame()

            finally:
                # Clean up the pending request
                if request_id in self._pending_requests:
                    del self._pending_requests[request_id]

        except Exception as e:
            logger.exception(f"{self} error generating TTS: {e}")
            error_message = f"TTS generation error: {str(e)}"
            yield ErrorFrame(error=error_message)

    def _convert_to_bytes(self, audio_data) -> bytes:
        """Convert various audio data formats to bytes."""
        if isinstance(audio_data, (bytes, bytearray)):
            return bytes(audio_data)

        # Handle numpy arrays
        try:
            import numpy as np

            if isinstance(audio_data, np.ndarray):
                # Ensure it's in the right format (16-bit PCM)
                if audio_data.dtype in [np.float32, np.float64]:
                    # Convert float [-1, 1] to int16 [-32768, 32767]
                    audio_data = np.clip(audio_data, -1.0, 1.0)  # Ensure values are in range
                    audio_data = (audio_data * 32767).astype(np.int16)
                elif audio_data.dtype != np.int16:
                    # Convert other integer types to int16
                    audio_data = audio_data.astype(np.int16)
                return audio_data.tobytes()
            elif hasattr(audio_data, 'tobytes'):
                return audio_data.tobytes()
            else:
                return bytes(audio_data)
        except ImportError:
            # Fallback if numpy is not available
            if hasattr(audio_data, 'tobytes'):
                return audio_data.tobytes()
            else:
                return bytes(audio_data)


class NeMoFastPitchHiFiGANTTSService(BaseNemoTTSService):
    """Text-to-Speech service using NeMo FastPitch-Hifigan model.

    More info: https://huggingface.co/nvidia/tts_en_fastpitch

    Args:
        fastpitch_model: FastPitch model name
        hifigan_model: Hifigan model name
        device: Device to run on (default: 'cuda')
        **kwargs: Additional arguments passed to BaseNemoTTSService
    """

    def __init__(
        self,
        fastpitch_model: str = "nvidia/tts_en_fastpitch",
        hifigan_model: str = "nvidia/tts_hifigan",
        device: str = "cuda",
        **kwargs,
    ):
        model_name = f"{fastpitch_model}+{hifigan_model}"
        self._fastpitch_model_name = fastpitch_model
        self._hifigan_model_name = hifigan_model
        super().__init__(model=model_name, device=device, **kwargs)

    def _setup_model(self):
        print("Loading model...")
        self._fastpitch_model = self._setup_fastpitch_model(self._fastpitch_model_name)
        self._hifigan_model = self._setup_hifigan_model(self._hifigan_model_name)
        return self._fastpitch_model, self._hifigan_model

    def _setup_fastpitch_model(self, model_name: str):
        if model_name.endswith(".nemo"):
            fastpitch_model = FastPitchModel.restore_from(model_name, map_location=torch.device(self._device))
        else:
            fastpitch_model = FastPitchModel.from_pretrained(model_name, map_location=torch.device(self._device))
        fastpitch_model.eval()
        return fastpitch_model

    def _setup_hifigan_model(self, model_name: str):
        if model_name.endswith(".nemo"):
            hifigan_model = HifiGanModel.restore_from(model_name, map_location=torch.device(self._device))
        else:
            hifigan_model = HifiGanModel.from_pretrained(model_name, map_location=torch.device(self._device))
        hifigan_model.eval()
        return hifigan_model

    def _generate_audio(self, text: str) -> Iterator[np.ndarray]:
        with torch.no_grad():
            parsed = self._fastpitch_model.parse(text)
            spectrogram = self._fastpitch_model.generate_spectrogram(tokens=parsed)
            audio = self._hifigan_model.convert_spectrogram_to_audio(spec=spectrogram)
            audio = audio.detach().view(-1).cpu().numpy()
            yield audio


class KokoroTTSService(BaseNemoTTSService, ToolCallingMixin):
    """Text-to-Speech service using Kokoro-82M model.

    Kokoro is an open-weight TTS model with 82 million parameters.
    More info: https://huggingface.co/hexgrad/Kokoro-82M

    Args:
        lang_code: Language code for the model (default: 'a' for American English)
        voice: Voice to use (default: 'af_heart')
        device: Device to run on (default: 'cuda')
        sample_rate: Audio sample rate in Hz (default: 24000 for Kokoro)
        **kwargs: Additional arguments passed to BaseNemoTTSService
    """

    def __init__(
        self,
        lang_code: str = "a",
        voice: str = "af_heart",
        device: str = "cuda",
        sample_rate: int = 24000,
        speed: float = 1.0,
        **kwargs,
    ):
        self._lang_code = lang_code
        self._voice = voice
        self._speed = speed
        assert speed > 0, "Speed must be greater than 0"
        model_name = f"kokoro-{lang_code}-{voice}"
        self._speed_lambda = 1.0
        self._original_speed = speed
        self._original_voice = voice
        self._gender = 'female' if voice[1] == 'f' else 'male'
        self._original_gender = self._gender
        self._original_lang_code = self._lang_code
        super().__init__(model=model_name, device=device, sample_rate=sample_rate, **kwargs)
        self.setup_tool_calling()

    def _setup_model(self, lang_code: Optional[str] = None, voice: Optional[str] = None):
        """Initialize the Kokoro pipeline."""
        try:
            from kokoro import KPipeline
        except ImportError:
            raise ImportError(
                "kokoro package is required for KokoroTTSService. " "Install it with: pip install kokoro>=0.9.2"
            )
        if lang_code is None:
            lang_code = self._lang_code
        if voice is None:
            voice = self._voice
        logger.info(f"Loading Kokoro TTS model with lang_code={lang_code}, voice={voice}")
        pipeline = KPipeline(lang_code=lang_code)
        return pipeline

    def _generate_audio(self, text: str) -> Iterator[np.ndarray]:
        """Generate audio using the Kokoro pipeline.

        Args:
            text: Text to convert to speech

        Yields:
            Audio data as numpy arrays
        """
        try:
            # Generate audio using Kokoro pipeline
            generator = self._model(text, voice=self._voice, speed=self._speed)

            # The generator yields tuples of (gs, ps, audio)
            # We only need the audio component
            for i, (gs, ps, audio) in enumerate(generator):
                logger.debug(
                    f"Kokoro generated audio chunk {i}: gs={gs}, ps={ps},"
                    f"audio_shape={audio.shape if hasattr(audio, 'shape') else len(audio)}"
                )
                if isinstance(audio, torch.Tensor):
                    audio = audio.detach().cpu().numpy()
                # Kokoro returns audio as numpy array in float32 format [-1, 1]
                # The base class will handle conversion to int16 bytes
                yield audio

        except Exception as e:
            logger.error(f"Error generating audio with Kokoro: {e}")
            raise

    async def tool_tts_set_speed(self, params: FunctionCallParams, speed_lambda: float):
        """
        Set a specific speaking speed of the assistant's voice.
        This tool should be called only when the user specifies the speed explicitly,
        such as "speak twice as fast" or "speak half as slow" or "speak 1.5 times as fast".

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.

        Args:
            speed_lambda: positive float, the relative change of the speaking speed to the original speed.
                        E.g., 1.0 for original speed, 1.25 for 25% faster than original speed,
                        0.8 for 20% slower than original speed.

        """
        if speed_lambda <= 0:
            result = {
                "success": False,
                "message": f"Speed remains unchanged since the change is not a positive number: {speed_lambda}",
            }
            logger.debug(f"Speed remains unchanged since the change is not a positive number: {speed_lambda}")
        else:
            self._speed = speed_lambda * self._original_speed
            result = {
                "success": True,
                "message": f"Speed set to {speed_lambda} of the original speed",
            }
            logger.debug(f"Speed set to {speed_lambda} of the original speed {self._original_speed}")
        await params.result_callback(result)

    async def tool_tts_reset_speed(self, params: FunctionCallParams):
        """
        Reset the speaking speed to the original speed.

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.
        """
        self._speed = self._original_speed
        result = {"success": True, "message": "Speaking speed is reset to the original one"}
        logger.debug(f"Speaking speed is reset to the original speed {self._original_speed}")
        await params.result_callback(result)

    async def tool_tts_speak_faster(self, params: FunctionCallParams):
        """
        Speak faster by increasing the speaking speed 15% faster each time this function is called.

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.
        """
        self._speed_lambda = self._speed_lambda + 0.15
        self._speed = self._speed_lambda * self._original_speed
        result = {
            "success": True,
            "message": f"Speaking speed is increased to {self._speed_lambda} of the original speed",
        }
        logger.debug(f"Speed is set to {self._speed_lambda} of the original speed {self._original_speed}")
        await params.result_callback(result)

    async def tool_tts_speak_slower(self, params: FunctionCallParams):
        """
        Speak slower by decreasing the speaking speed 15% slower each time this function is called.

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.
        """
        self._speed_lambda = self._speed_lambda - 0.15
        if self._speed_lambda < 0.1:
            self._speed = 0.1 * self._original_speed
            result = {
                "success": True,
                "message": "Speaking speed is decreased to the minimum of 0.1 of the original speed",
            }
            logger.debug(f"Speed is set to the minimum of 0.1 of the original speed {self._original_speed}")
        else:
            self._speed = self._speed_lambda * self._original_speed
            result = {
                "success": True,
                "message": f"Speaking speed is decreased to {self._speed_lambda} of the original speed",
            }
            logger.debug(f"Speed is set to {self._speed_lambda} of the original speed {self._original_speed}")
        await params.result_callback(result)

    async def tool_tts_set_voice(self, params: FunctionCallParams, accent: str, gender: str):
        """
        Set the accent and gender of the assistant's voice.
        This tool should be called only when the user specifies the accent and/or gender explicitly.

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.

        Args:
            accent: Accent for the TTS model. Must be one of 'American English', 'British English'
                    or 'current' for keeping the current accent.
            gender: gender of the assistant's voice. Must be one of 'male', 'female',
                    or 'current' for keeping the current gender.
        """
        await params.llm.push_frame(LLMTextFrame("Just a moment."))

        lang_code = "a" if accent == "American English" else "b" if accent == "British English" else "current"
        new_lang_code = self._lang_code
        new_gender = self._gender
        if lang_code != 'current':
            new_lang_code = lang_code
        if gender != 'current':
            new_gender = gender

        if new_lang_code == 'a':
            new_voice = 'af_heart' if new_gender == 'female' else 'am_michael'
        elif new_lang_code == 'b':
            new_voice = 'bf_emma' if new_gender == 'female' else 'bm_george'
        else:
            await params.result_callback(
                {
                    "success": False,
                    "message": f"Invalid language code: {new_lang_code} or gender: {new_gender}",
                }
            )
            return

        new_model = await asyncio.to_thread(self._setup_model, new_lang_code, new_voice)
        self._model = new_model
        self._lang_code = new_lang_code
        self._gender = new_gender
        self._voice = new_voice
        logger.debug(f"Language and voice are set to {new_lang_code} and {new_voice}")
        await params.result_callback({"success": True, "message": "Done. Language and voice are set to the new ones."})

    async def tool_tts_reset_voice(self, params: FunctionCallParams):
        """
        Reset the accent and voice to the original ones.

        After calling this tool, continue the previous response if it was unfinished and was
        interrupted by calling this tool, otherwise start a new response.
        """
        await params.llm.push_frame(LLMTextFrame("Of course."))

        new_model = await asyncio.to_thread(self._setup_model, self._original_lang_code, self._original_voice)
        self._model = new_model
        self._lang_code = self._original_lang_code
        self._gender = self._original_gender
        self._voice = self._original_voice
        logger.debug(
            f"Language and voice are reset to the original ones {self._original_lang_code} and {self._original_voice}"
        )
        await params.result_callback(
            {"success": True, "message": "Done. Language and voice are reset to the original ones."}
        )

    def setup_tool_calling(self):
        """
        Setup the tool calling mixin by registering all available tools.
        """
        self.register_direct_function("tool_tts_reset_speed", self.tool_tts_reset_speed)
        self.register_direct_function("tool_tts_speak_faster", self.tool_tts_speak_faster)
        self.register_direct_function("tool_tts_speak_slower", self.tool_tts_speak_slower)
        self.register_direct_function("tool_tts_set_speed", self.tool_tts_set_speed)
        self.register_direct_function("tool_tts_set_voice", self.tool_tts_set_voice)
        self.register_direct_function("tool_tts_reset_voice", self.tool_tts_reset_voice)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Callable, TypeVar

import fiddle as fdl

T = TypeVar("T")


class PredicateDispatch:
    """A dispatcher that routes values to handlers based on predicates.

    This class implements a predicate-based dispatch system where handlers are registered
    with conditions that determine when they should be used. When called, it tries each
    predicate in order until it finds a match.

    Example:
        ```python
        dispatcher = PredicateDispatch()

        @dispatcher.register(lambda x: isinstance(x, str))
        def handle_strings(s):
            return f"String: {s}"

        result = dispatcher("hello")  # Returns "String: hello"
        ```
    """

    def __init__(self):
        """Initialize an empty handler registry."""
        self.handlers = []

    def register(self, predicate: Callable[[Any], bool]):
        """Register a new handler with a predicate.

        Args:
            predicate: A callable that takes a value and returns True if the handler
                should process this value.

        Returns:
            A decorator function that registers the handler.
        """

        def decorator(func: Callable[[T], Any]):
            self.handlers.append((predicate, func))
            return func

        return decorator

    def __call__(self, value):
        """Process a value through the registered handlers.

        Args:
            value: The value to be processed.

        Returns:
            The processed value from the first matching handler, or the original value
            if no handlers match.
        """
        for predicate, handler in self.handlers:
            if predicate(value):
                return handler(value)
        return value  # default case: return unchanged

    def register_class(self, cls: type):
        """Register a handler for instances of a specific class.

        A convenience method that automatically creates an isinstance predicate.

        Args:
            cls: The class to check instances against.

        Returns:
            A decorator function that registers the handler.
        """
        return self.register(lambda v: isinstance(v, cls))


"""Global dispatcher for converting Python objects to Fiddle configurations.

This dispatcher is used by Fiddle's serialization system to handle special cases
during configuration serialization. When Fiddle encounters an object it doesn't
know how to serialize, it will pass it through this dispatcher to convert it
into a serializable Fiddle configuration.

Example use cases:
    - Converting functools.partial to fdl.Partial
    - Converting HuggingFace models to their from_pretrained configurations
    - Handling custom classes with special serialization needs

The dispatcher is extended by registering new handlers with predicates that
determine when they should be used. See PredicateDispatch for more details.
"""
to_config = PredicateDispatch()


@to_config.register_class(partial)
def handle_partial(value: partial):
    """Convert functools.partial objects to Fiddle Partial configurations.

    This handler enables serialization of partial function applications by converting
    them to Fiddle's equivalent representation.

    Args:
        value: A functools.partial object.

    Returns:
        A Fiddle Partial configuration representing the same partial application.
    """
    return fdl.Partial(value.func, *value.args, **value.keywords)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import dataclasses
import re
from collections import defaultdict
from enum import Enum, auto
from io import BytesIO
from typing import Any, List, Optional, Union

from PIL import Image
from transformers import AutoTokenizer


class SeparatorStyle(Enum):
    """Different separator style."""

    SINGLE = auto()
    TWO = auto()
    MPT = auto()
    PLAIN = auto()
    CHATML = auto()
    LLAMA_2 = auto()
    LLAMA_3 = auto()
    MLLAMA = auto()
    MISTRAL = auto()
    NVGPT = auto()
    QWEN = auto()
    GEMMA = auto()


@dataclasses.dataclass
class Conversation:
    """A class that keeps all conversation history."""

    system: Optional[str]
    roles: tuple[str, str]
    messages: List[List[str]]
    offset: int
    sep_style: SeparatorStyle = SeparatorStyle.SINGLE
    sep: str = "###"
    sep2: str = None
    version: str = "Unknown"

    tokenizer_name_or_path: Any = None
    stop_str: Union[str, List[str]] = None
    stop_token_ids: List[int] = None

    skip_next: bool = False

    def process_prompt_with_images(self, messages):
        # Process messages to handle potential image tokens.
        return messages

    def process_chat_template(self, tokenizer_name_or_path, messages):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)
        if self.system is None:
            chat = []
        else:
            chat = [{"role": "system", "content": self.system}]
        for role, message in messages:
            chat.append({"role": role.lower(), "content": message})
        ret = tokenizer.apply_chat_template(chat, tokenize=False, add_generation_prompt=False)
        return ret

    def get_prompt(self):
        messages = self.messages

        if self.sep_style == SeparatorStyle.SINGLE:
            ret = self.system + self.sep
            for role, message in messages:
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += role + ": " + message + self.sep
                else:
                    ret += role + ":"

        elif self.sep_style == SeparatorStyle.TWO:
            """
            A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. USER: {{ user_message_1 }} ASSISTANT: {{ model_answer_1 }}</s>USER: {{ user_message_2 }}
            """
            seps = [self.sep, self.sep2]
            ret = self.system + seps[0]
            for i, (role, message) in enumerate(messages):
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += role + ": " + message + seps[i % 2]
                    # Add space to make sure the labels can be correctly generated.
                    self.messages[i][1] = " " + self.messages[i][1]
                else:
                    ret += role + ":"

        elif self.sep_style == SeparatorStyle.MISTRAL and self.version == "vila":
            """
            <s>[INST] {{ user_message_1 }} [/INST]{{ model_answer_1 }}</s>[INST] {{ user_message_2 }} [/INST]
            """
            wrap_sys = lambda msg: f"{msg}" + ("\n" if msg else "")
            wrap_inst = lambda msg: f"[INST] {msg} [/INST]"
            ret = "<s>"

            for i, (role, message) in enumerate(messages):
                if i == 0:
                    assert message, "first message should not be none"
                    assert role == self.roles[0], "first message should come from user"
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    if i == 0:
                        message = wrap_sys(self.system) + message
                    if i % 2 == 0:
                        message = wrap_inst(message)
                        ret += self.sep + message
                    else:
                        ret += message + self.sep2
                else:
                    ret += ""

        elif self.sep_style == SeparatorStyle.LLAMA_2:
            """
            <s>[INST] <<SYS>>
            You are a helpful language and vision assistant. You are able to understand the visual content that the user provides, and assist the user with a variety of tasks using natural language.
            <</SYS>>

            {{ user_message_1 }} [/INST] {{ model_answer_1 }} </s><s>[INST] {{ user_message_2 }} [/INST]
            """
            tokenizer_name_or_path = self.tokenizer_name_or_path or "meta-llama/Llama-2-7b-chat-hf"
            ret = self.process_chat_template(tokenizer_name_or_path, messages)

        elif self.sep_style == SeparatorStyle.LLAMA_3:
            """
            <|begin_of_text|><|start_header_id|>system<|end_header_id|>

            {{ system_prompt }}<|eot_id|><|start_header_id|>user<|end_header_id|>

            {{ user_message_1 }}<|eot_id|><|start_header_id|>assistant<|end_header_id|>

            {{ model_answer_1 }}<|eot_id|><|start_header_id|>user<|end_header_id|>

            {{ user_message_2 }}<|eot_id|><|start_header_id|>assistant<|end_header_id|>
            """
            tokenizer_name_or_path = self.tokenizer_name_or_path or "meta-llama/Meta-Llama-3-8B-Instruct"
            ret = self.process_chat_template(tokenizer_name_or_path, messages)

        elif self.sep_style == SeparatorStyle.MLLAMA:
            tokenizer_name_or_path = self.tokenizer_name_or_path or "meta-llama/Llama-3.2-11B-Vision-Instruct"
            ret = self.process_chat_template(tokenizer_name_or_path, messages)

        elif self.sep_style == SeparatorStyle.NVGPT:
            ret = self.sep2 + self.system + self.sep
            for role, message in messages:
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += role + '\n' + message + '\n' + self.sep
                else:
                    ret += role + '\n'

        elif self.sep_style == SeparatorStyle.PLAIN:
            seps = [self.sep, self.sep2]
            ret = self.system
            for i, (role, message) in enumerate(messages):
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += message + seps[i % 2]
                else:
                    ret += ""

        elif self.sep_style == SeparatorStyle.MISTRAL:
            """
            NOT tested in NeMo!
            """
            tokenizer_name_or_path = self.tokenizer_name_or_path or "mistralai/Mistral-7B-Instruct-v0.2"
            ret = self.process_chat_template(tokenizer_name_or_path, messages)

        elif self.sep_style == SeparatorStyle.CHATML:
            """
            NOT tested in NeMo!
            """
            ret = "" if self.system == "" else self.system + self.sep + "\n"
            for role, message in messages:
                if message:
                    if type(message) is tuple:
                        message, images = message
                        message = "<image>" * len(images) + message
                    ret += role + "\n" + message + self.sep + "\n"
                else:
                    ret += role + "\n"
            return ret

        elif self.sep_style == SeparatorStyle.MPT:
            """
            NOT tested in NeMo!
            """
            ret = self.system + self.sep
            for role, message in messages:
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += role + message + self.sep
                else:
                    ret += role

        elif self.sep_style == SeparatorStyle.GEMMA:
            """
            NOT tested in NeMo!
            """
            ret = ""
            for i, (role, message) in enumerate(messages):
                assert role == self.roles[i % 2], "Conversation should alternate user/assistant/user/assistant/..."
                if message:
                    if type(message) is tuple:
                        message, _, _ = message
                    ret += role + message + self.sep
                else:
                    ret += role

        else:
            raise ValueError(f"Invalid style: {self.sep_style}")

        return ret

    def append_message(self, role, message):
        self.messages.append([role, message])

    def process_image(self, image, image_process_mode, return_pil=False, image_format="PNG"):
        if image_process_mode == "Pad":

            def expand2square(pil_img, background_color=(122, 116, 104)):
                width, height = pil_img.size
                if width == height:
                    return pil_img
                elif width > height:
                    result = Image.new(pil_img.mode, (width, width), background_color)
                    result.paste(pil_img, (0, (width - height) // 2))
                    return result
                else:
                    result = Image.new(pil_img.mode, (height, height), background_color)
                    result.paste(pil_img, ((height - width) // 2, 0))
                    return result

            image = expand2square(image)
        elif image_process_mode in ["Default", "Crop"]:
            pass
        elif image_process_mode == "Resize":
            image = image.resize((336, 336))
        else:
            raise ValueError(f"Invalid image_process_mode: {image_process_mode}")

        if type(image) is not Image.Image:
            image = Image.open(image).convert("RGB")

        max_hw, min_hw = max(image.size), min(image.size)
        aspect_ratio = max_hw / min_hw
        max_len, min_len = 1008, 672
        shortest_edge = int(min(max_len / aspect_ratio, min_len, min_hw))
        longest_edge = int(shortest_edge * aspect_ratio)
        W, H = image.size
        if H > W:
            H, W = longest_edge, shortest_edge
        else:
            H, W = shortest_edge, longest_edge
        image = image.resize((W, H))
        if return_pil:
            return image
        else:
            buffered = BytesIO()
            image.save(buffered, format=image_format)
            img_b64_str = base64.b64encode(buffered.getvalue()).decode()
            return img_b64_str

    def get_images(self, return_pil=False, return_path=False):
        images = []
        for i, (role, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                if type(msg) is tuple:
                    msg, image, image_process_mode = msg
                    if type(image) != list:
                        image = [image]
                    for img in image:
                        if not return_path:
                            img = self.process_image(img, image_process_mode, return_pil=return_pil)
                        images.append(img)
        return images

    def to_gradio_chatbot(self):
        ret = []
        for i, (role, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                if type(msg) is tuple:
                    msg, image, image_process_mode = msg
                    if type(image) != list:
                        image = [image]
                    if len(image) == 1:
                        msg = "<image>\n" + msg.replace("<image>", "").strip()
                    else:
                        msg = re.sub(r"(<image>)\n(?=<image>)", r"\1 ", msg)
                    for img in image:
                        img_b64_str = self.process_image(img, "Default", return_pil=False, image_format="JPEG")
                        img_str = f'<img src="data:image/jpeg;base64,{img_b64_str}"/>'
                        msg = msg.replace("<image>", img_str, 1).strip()
                    if len(msg) > 0:
                        ret.append([msg, None])
                else:
                    ret.append([msg, None])
            else:
                ret[-1][-1] = msg
        return ret

    def copy(self):
        return Conversation(
            system=self.system,
            roles=self.roles,
            messages=[[x, y] for x, y in self.messages],
            offset=self.offset,
            sep_style=self.sep_style,
            sep=self.sep,
            sep2=self.sep2,
            version=self.version,
        )

    def dict(self):
        if len(self.get_images()) > 0:
            return {
                "system": self.system,
                "roles": self.roles,
                "messages": [[x, y[0] if type(y) is tuple else y] for x, y in self.messages],
                "offset": self.offset,
                "sep": self.sep,
                "sep2": self.sep2,
            }
        return {
            "system": self.system,
            "roles": self.roles,
            "messages": self.messages,
            "offset": self.offset,
            "sep": self.sep,
            "sep2": self.sep2,
        }


# Conversation Template for NVGPT
conv_nvgpt = Conversation(
    system="""A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.\n\n""",
    roles=("User", "Assistant"),
    version="nvgpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.NVGPT,
    sep="<extra_id_1>",
    sep2=f"<extra_id_0>System\n",
)

conv_nv_dpo = Conversation(
    system="\n",
    roles=("User", "Assistant"),
    version="nv_dpo",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.NVGPT,
    sep="<extra_id_1>",
    sep2=f"<extra_id_0>System\n",
)

conv_vicuna_v0 = Conversation(
    system="A chat between a curious human and an artificial intelligence assistant. "
    "The assistant gives helpful, detailed, and polite answers to the human's questions.",
    roles=("Human", "Assistant"),
    messages=[
        ["Human", "What are the key differences between renewable and non-renewable energy sources?"],
        [
            "Assistant",
            "Renewable energy sources are those that can be replenished naturally in a relatively "
            "short amount of time, such as solar, wind, hydro, geothermal, and biomass. "
            "Non-renewable energy sources, on the other hand, are finite and will eventually be "
            "depleted, such as coal, oil, and natural gas. Here are some key differences between "
            "renewable and non-renewable energy sources:\n"
            "1. Availability: Renewable energy sources are virtually inexhaustible, while non-renewable "
            "energy sources are finite and will eventually run out.\n"
            "2. Environmental impact: Renewable energy sources have a much lower environmental impact "
            "than non-renewable sources, which can lead to air and water pollution, greenhouse gas emissions, "
            "and other negative effects.\n"
            "3. Cost: Renewable energy sources can be more expensive to initially set up, but they typically "
            "have lower operational costs than non-renewable sources.\n"
            "4. Reliability: Renewable energy sources are often more reliable and can be used in more remote "
            "locations than non-renewable sources.\n"
            "5. Flexibility: Renewable energy sources are often more flexible and can be adapted to different "
            "situations and needs, while non-renewable sources are more rigid and inflexible.\n"
            "6. Sustainability: Renewable energy sources are more sustainable over the long term, while "
            "non-renewable sources are not, and their depletion can lead to economic and social instability.\n",
        ],
    ],
    offset=2,
    sep_style=SeparatorStyle.SINGLE,
    sep="###",
)

conv_vicuna_v1 = Conversation(
    system="A chat between a curious user and an artificial intelligence assistant. "
    "The assistant gives helpful, detailed, and polite answers to the user's questions.",
    roles=("USER", "ASSISTANT"),
    version="v1",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.TWO,
    sep=" ",
    sep2="</s>",
    stop_str="</s>",
)

conv_llama_2 = Conversation(
    system="""You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe.  Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.

If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information.""",
    roles=("USER", "ASSISTANT"),
    version="llama_v2",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.LLAMA_2,
    sep="<s>",
    sep2="</s>",
    stop_str=" </s>",
)

conv_llava_llama_2 = Conversation(
    system="You are a helpful language and vision assistant. "
    "You are able to understand the visual content that the user provides, "
    "and assist the user with a variety of tasks using natural language.",
    roles=("USER", "ASSISTANT"),
    version="llama_v2",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.LLAMA_2,
    sep="<s>",
    sep2="</s>",
    stop_str=" </s>",
)

conv_llava_llama_3 = Conversation(
    system="You are a helpful language and vision assistant. "
    "You are able to understand the visual content that the user provides, "
    "and assist the user with a variety of tasks using natural language.",
    roles=("user", "assistant"),
    version="llama_v3",
    messages=[],
    offset=0,
    sep="<|eot_id|>",
    sep_style=SeparatorStyle.LLAMA_3,
    tokenizer_name_or_path="meta-llama/Meta-Llama-3-8B-Instruct",
    stop_str="<|eot_id|>",
)

conv_mllama = Conversation(
    system="",
    roles=("user", "assistant"),
    version="llama_v3_2",
    messages=[],
    offset=0,
    sep="<|eot_id|>",
    sep_style=SeparatorStyle.MLLAMA,
    tokenizer_name_or_path="meta-llama/Llama-3.2-11B-Vision-Instruct",
    stop_str="<|eot_id|>",
)

conv_mistral_instruct = Conversation(
    system="",
    roles=("USER", "ASSISTANT"),
    version="llama_v2",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.LLAMA_2,
    sep="",
    sep2="</s>",
    stop_str=" </s>",
)

conv_llava_llama_2_simple = Conversation(
    system="Answer the questions about the visual content that the user provides.",
    roles=("USER", "ASSISTANT"),
    version="llama_v2",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.LLAMA_2,
    sep="<s>",
    sep2="</s>",
    stop_str=" </s>",
)

conv_llava_llama_2_mmtag = Conversation(
    system="Answer the questions about the visual content that the user provides."
    "The visual content will be provided with the following format: <Image>visual content</Image>.",
    roles=("USER", "ASSISTANT"),
    version="llama_v2_mmtag",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.LLAMA_2,
    sep="<s>",
    sep2="</s>",
    stop_str=" </s>",
)

conv_mpt = Conversation(
    system="""<|im_start|>system
A conversation between a user and an LLM-based AI assistant. The assistant gives helpful and honest answers.""",
    roles=("<|im_start|>user\n", "<|im_start|>assistant\n"),
    version="mpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MPT,
    sep="<|im_end|>",
)

conv_qwen = Conversation(
    system="""<|im_start|>system
You are a helpful assistant.""",
    roles=("<|im_start|>user", "<|im_start|>assistant"),
    version="qwen",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.CHATML,
    sep="<|im_end|>",
)

conv_gemma_instruct = Conversation(
    system="",
    roles=("<start_of_turn>user\n", "<start_of_turn>model\n"),
    version="gemma",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.GEMMA,
    sep="<end_of_turn>\n",
)

conv_llava_plain = Conversation(
    system="",
    roles=("", ""),
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.PLAIN,
    sep="",
    sep2="\n",
    stop_str="\n",
)

conv_llava_v0 = Conversation(
    system="A chat between a curious human and an artificial intelligence assistant. "
    "The assistant gives helpful, detailed, and polite answers to the human's questions.",
    roles=("Human", "Assistant"),
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.SINGLE,
    sep="###",
)

conv_llava_v0_mmtag = Conversation(
    system="A chat between a curious user and an artificial intelligence assistant. "
    "The assistant is able to understand the visual content that the user provides, and assist the user with a variety of tasks using natural language."
    "The visual content will be provided with the following format: <Image>visual content</Image>.",
    roles=("Human", "Assistant"),
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.SINGLE,
    sep="###",
    version="v0_mmtag",
)

conv_llava_v1 = Conversation(
    system="A chat between a curious human and an artificial intelligence assistant. "
    "The assistant gives helpful, detailed, and polite answers to the human's questions.",
    roles=("USER", "ASSISTANT"),
    version="v1",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.TWO,
    sep=" ",
    sep2="</s>",
)

conv_llava_v1_mmtag = Conversation(
    system="A chat between a curious user and an artificial intelligence assistant. "
    "The assistant is able to understand the visual content that the user provides, and assist the user with a variety of tasks using natural language."
    "The visual content will be provided with the following format: <Image>visual content</Image>.",
    roles=("USER", "ASSISTANT"),
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.TWO,
    sep=" ",
    sep2="</s>",
    version="v1_mmtag",
)

conv_mistral_vila = Conversation(
    system=None,
    roles=("USER", "ASSISTANT"),
    version="vila",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MISTRAL,
    sep="",
    sep2="</s>",
    stop_str="</s>",
)

conv_mistral_orca = Conversation(
    system="""<|im_start|>system
You are MistralOrca, a large language model trained by Alignment Lab AI. Write out your reasoning step-by-step to be sure you get the right answers!""",
    roles=("<|im_start|>user\n", "<|im_start|>assistant\n"),
    version="mpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MPT,
    sep="<|im_end|>",
)

conv_mistral_zephyr = Conversation(
    system="""<|system|>
You are a helpful AI assistant.""",
    roles=("<|user|>\n", "<|assistant|>\n"),
    version="mpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MPT,
    sep="</s>",
)

conv_mistral_direct = Conversation(
    system="""<|im_start|>system
Answer the questions.""",
    roles=("<|im_start|>user\n", "<|im_start|>assistant\n"),
    version="mpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MPT,
    sep="<|im_end|>",
)

conv_chatml_direct = Conversation(
    system="""<|im_start|>system
Answer the questions.""",
    roles=("<|im_start|>user\n", "<|im_start|>assistant\n"),
    version="mpt",
    messages=[],
    offset=0,
    sep_style=SeparatorStyle.MPT,
    sep="<|im_end|>",
)

default_conversation = conv_vicuna_v1
conv_templates = {
    "default": conv_vicuna_v1,
    "v0": conv_vicuna_v0,
    "v1": conv_vicuna_v1,
    "vicuna_v1": conv_vicuna_v1,
    "llama_2": conv_llama_2,
    "mllama": conv_mllama,
    "mistral_instruct": conv_mistral_instruct,
    "mistral_orca": conv_mistral_orca,
    "mistral_zephyr": conv_mistral_zephyr,
    "mistral_direct": conv_mistral_direct,
    "mistral": conv_mistral_vila,
    "plain": conv_llava_plain,
    "v0_plain": conv_llava_plain,
    "chatml_direct": conv_chatml_direct,
    "llava_v0": conv_llava_v0,
    "llava_v0_mmtag": conv_llava_v0_mmtag,
    "llava_v1": conv_llava_v1,
    "llava_v1_mmtag": conv_llava_v1_mmtag,
    "llava_llama_2": conv_llava_llama_2,
    "llava_llama_3": conv_llava_llama_3,
    "llava_llama_2_simple": conv_llava_llama_2_simple,
    "llava_llama_2_mmtag": conv_llava_llama_2_mmtag,
    "llava_mistral_instruct": conv_mistral_instruct,
    "mpt": conv_mpt,
    "qwen_1_5": conv_qwen,
    "gemma_instruct": conv_gemma_instruct,
    "nvgpt": conv_nvgpt,
    "nv_steerlm": conv_nvgpt,
    "nv_dpo": conv_nv_dpo,
}

if __name__ == "__main__":
    print(default_conversation.get_prompt())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import nemo_run as run

from nemo.collections import llm
from nemo.collections.llm.modelopt import PruningConfig

NAME = "default"


@run.cli.factory(target=llm.prune, name=NAME)
def prune_recipe(nemo_checkpoint: str, save_path: str) -> run.Partial:
    # pylint: disable=line-too-long
    """Create a generic pruning recipe.

    This function sets up a complete configuration for pruning, including trainer and data.

    Args:
        nemo_checkpoint (str): The path to the NeMo checkpoint to be pruned.
        save_path (str): The path to save the pruned NeMo checkpoint.

    Returns:
        run.Partial: Partial configuration for pruning.

    Python recipe usage:
        >>> import nemo_run as run
        >>> from nemo.collections import llm
        >>> from nemo.collections.llm.modelopt.recipes import prune_recipe
        >>> recipe = prune_recipe(
        ...     nemo_checkpoint="/path/to/llama3.1-8b/nemo-ckpt/",
        ...     save_path="/path/to/pruned/llama3.1-8b/nemo-ckpt/",
        ... )
        >>> recipe.devices = 8
        >>> recipe.pp_size = 8
        >>> recipe.data = run.Config(
        ...     llm.PreTrainingDataModule,
        ...     paths=["1.0", "path/to/tokenized/data"],
        ...     seq_length=8192,
        ...     micro_batch_size=1,
        ...     global_batch_size=1,  # should be equal to micro_batch_size
        ... )
        >>> recipe.pruning_config.target_ffn_hidden_size = 9216
        >>> recipe.pruning_config.target_hidden_size = 3072
        >>> ...
        >>> run.run(recipe)
    """
    recipe = run.Partial(
        llm.prune,
        nemo_checkpoint=nemo_checkpoint,
        save_path=save_path,
        pruning_config=run.Config(PruningConfig),
        devices=1,
        num_nodes=1,
        tp_size=1,
        pp_size=1,
        num_train_samples=1024,
        data=run.Config(llm.MockDataModule, seq_length=256, micro_batch_size=1, global_batch_size=1),
        legacy_ckpt=False,
    )

    return recipe
